/* hp2100_di.h: HP 12821A HP-IB Disc Interface simulator declarations

   Copyright (c) 2010-2022, J. David Bryan

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of the author shall not be
   used in advertising or otherwise to promote the sale, use or other dealings
   in this Software without prior written authorization from the author.

   DI           12821A Disc Interface

   27-Mar-22    JDB     Added "da_bus_initialize" external
   11-Mar-22    JDB     Added "di_initializer" external
   18-Oct-21    JDB     Added "other" field to DI_STATE
   13-Jul-21    JDB     Added "di_bus_trace" global and tracing macros
   25-Jun-21    JDB     Removed "di_poll_response"; now handled by devices
   16-Jun-21    JDB     Added bus INITIALIZER definition
   18-May-21    JDB     Logically separated SIMH unit numbers from bus addresses
   04-May-21    JDB     Added dc_bus_accept, dc_bus_respond globals
   01-Apr-21    JDB     Removed REG_FIT from register definitions (obsolete)
   07-Feb-21    JDB     Changed TMR register macro from BRDATA to SVDATA
   11-Jul-18    JDB     Revised I/O model
   15-Mar-17    JDB     Trace flags are now global
   09-Mar-17    JDB     Changed the DIAG option to DIAGNOSTIC
   10-Jan-17    JDB     Moved byte accessors to hp2100_defs.h
   13-May-16    JDB     Modified for revised SCP API function parameter types
   14-Feb-12    JDB     First release
   16-Nov-10    JDB     Created DI common definitions file


   This file defines the interface between HP-IB device simulators and the
   12821A Disc Interface simulator.  It must be included by the device-specific
   modules (hp2100_di_da.c, etc.).


   Implementation notes:

    1. Three CARD_ID values are defined, corresponding to the Amigo disc (DA),
       CS/80 disc (DC), and Amigo mag tape (MA) simulators.  The MA device is
       reserved for future use; only the DA and DC devices are currently
       implemented.
*/



/* Bus tracing.

   Tracing bus transfers produces messages in this format:

     DIO | ATN | EOI | NRFD | IFC | HP-IB Mnemonic
     94H | ATN |     |      |     | Universal clear
     80H | ATN | EOI |      |     | Parallel poll
     20H |     |     |      |     | Data (device 0 accepts)

   To avoid the requirement that each trace call perform the bus signal
   formatting, two macros are provided to interface to the "di_bus_trace"
   routine, which provides the initial columns and then calls "hp_trace" to
   format the remainder of the trace line.  The macro invocation requires just
   the "HP-IB Mnemonic" information.  That information may involve numeric
   formatting, as the "device 0" message above shows, so the invocation passes a
   format string and optional parameters, just as with the standard "tprintf"
   macro.

   "tprintf_bus" is used to write bus tracing messages.  It calls
   "di_bus_format" if the debug stream is open and the TRACE_XFER option is
   currently enabled in device "dev".  Otherwise, it's a NOP.  "..." is the
   format string and associated values.

   "tpprintf_bus" is identical to "tprintf_bus", except that a device pointer is
   passed instead of a device structure.
*/

#define tprintf_bus(dev, signals, ...) \
          if (TRACING (dev, TRACE_XFER)) \
              di_bus_trace (&(dev), (signals), __VA_ARGS__); \
          else \
              (void) 0

#define tpprintf_bus(dptr, signals, ...) \
          if (TRACINGP (dptr, TRACE_XFER)) \
              di_bus_trace ((dptr), (signals), __VA_ARGS__); \
          else \
              (void) 0


/* Program constants */

#define FIFO_SIZE           16                  /* FIFO depth in words */

typedef enum {                                  /* interface card identifiers */
    da,                                         /*   Amigo disc interface */
    dc,                                         /*   CS/80 disc interface */
    ma,                                         /*   Amigo tape interface */
    Card_Count,                                 /*   count of card IDs */
    First_Card = da,                            /*   first card ID */
    Last_Card  = ma                             /*   last card ID */
    } CARD_ID;


/* HP-IB constants */

#define ADDR_COUNT          8                   /* count of valid bus addresses */
#define MAX_ADDR            (ADDR_COUNT - 1)    /* maximum bus address number */



/* Device flags and accessors.

   DEV_V_UF + 15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
             +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
             | - | - | P | D | R |  address  | reserved for disc/tape flags  |
             +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     P = DCPC pacing is in position A/B (0/1)
     D = diagnostic mode is enabled
     R = realtime mode is enabled
*/

#define DEV_BUSADR_SHIFT    (DEV_V_UF +  8)     /* interface HP-IB address for diagnostics */
#define DEV_REALTIME_SHIFT  (DEV_V_UF + 11)     /* timing mode is realistic */
#define DEV_DIAG_SHIFT      (DEV_V_UF + 12)     /* card is in diagnostic mode */
#define DEV_W1_SHIFT        (DEV_V_UF + 13)     /* DCPC pacing jumper */

#define DEV_BUSADR_MASK     07u                 /* bus address mask */

#define DEV_BUSADR          (DEV_BUSADR_MASK << DEV_BUSADR_SHIFT)

#define DEV_REALTIME        (1u << DEV_REALTIME_SHIFT)
#define DEV_DIAG            (1u << DEV_DIAG_SHIFT)
#define DEV_W1              (1u << DEV_W1_SHIFT)

#define DEV_ADDR(f)         (((f) >> DEV_BUSADR_SHIFT) & DEV_BUSADR_MASK)
#define TO_DEV_ADDR(a)      (((a) & DEV_BUSADR_MASK) << DEV_BUSADR_SHIFT)


/* HP-IB bus control and data lines.

   The HP-IB contains eight data lines, three handshake lines, and five
   bus management lines.  The state of the bus is represented by a 16-bit value,
   organized as follows:

      15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | S | I | R | N | V | D | E | A |           data byte           |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     S = SRQ (service request)
     I = IFC (interface clear)
     R = REN (remote enable)
     N = NRFD (not ready for data)
     V = DAV (data valid)
     D = NDAC (not data accepted)
     E = EOI (end or identify)
     A = ATN (attention)


   Implementation notes:

    1. The SRQ, IFC, REN, EOI, and ATN bits align directly with their
       counterparts in the status register, so they must not be moved.  EOI and
       ATN align with their counterparts in the FIFO tag register when
       left-shifted by eight bits.

    2. We supply type definitions for the bus control lines, data lines, and
       combined lines, but they are all interchangeably represented.
*/

typedef enum {                                  /* HP-IB control lines */
    Bus_ATN  = 0000400,                         /*   Attention */
    Bus_EOI  = 0001000,                         /*   End or Identify */
    Bus_NDAC = 0002000,                         /*   Not Data Accepted */
    Bus_DAV  = 0004000,                         /*   Data Valid */
    Bus_NRFD = 0010000,                         /*   Not Ready for Data */
    Bus_REN  = 0020000,                         /*   Remote Enable */
    Bus_IFC  = 0040000,                         /*   Interface Clear */
    Bus_SRQ  = 0100000                          /*   Service Request */
    } BUS_CNTL;

typedef BUS_CNTL            BUS_CABLE;          /* the sixteen control and data lines */
typedef BUS_CNTL            BUS_DATA;           /* the eight data lines */

#define BUS_CNTL_MASK       0177400u            /* mask for bus control signals */
#define BUS_DATA_MASK       0000377u            /* mask for bus data */

#define CNTL(b)             ((b) & BUS_CNTL_MASK)       /* control lines are in the upper byte */
#define DATA(b)             ((b) & BUS_DATA_MASK)       /* data lines are in the lower byte */

#define BUS_NONE            0                               /* no signals asserted */
#define BUS_PPOLL           (Bus_ATN | Bus_EOI)             /* signals for a parallel poll */
#define BUS_CNTLR           (Bus_ATN | Bus_REN | Bus_IFC)   /* signals reserved for the controller */


/* HP-IB command decoding */

#define BUS_ADDRESS_MASK    0037u               /* bus address mask */
#define BUS_SECONDARY_MASK  0037u               /* bus secondary mask */
#define BUS_GROUP_MASK      0140u               /* bus group mask */
#define BUS_COMMAND_MASK    0177u               /* bus command mask */

#define BUS_ADDRESS_SHIFT   0
#define BUS_SECONDARY_SHIFT 0
#define BUS_GROUP_SHIFT     5
#define BUS_COMMAND_SHIFT   0

#define BUS_ADDRESS(b)      (((b) & BUS_ADDRESS_MASK)   >> BUS_ADDRESS_SHIFT)
#define BUS_SECONDARY(b)    (((b) & BUS_SECONDARY_MASK) >> BUS_SECONDARY_SHIFT)
#define BUS_GROUP(b)        (((b) & BUS_GROUP_MASK)     >> BUS_GROUP_SHIFT)
#define BUS_COMMAND(b)      (((b) & BUS_COMMAND_MASK)   >> BUS_COMMAND_SHIFT)

#define BUS_PCG             BUS_GROUP (0000u)   /* primary command group */
#define BUS_LAG             BUS_GROUP (0040u)   /* listen address group */
#define BUS_TAG             BUS_GROUP (0100u)   /* talk address group */
#define BUS_SCG             BUS_GROUP (0140u)   /* secondary command group */

#define BUS_UNADDRESS       0037u               /* unlisten and untalk addresses */
#define BUS_UNTALK          0137u               /* untalk command (for Identify) */

#define PPR(a)              (1u << (7 - (a)))   /* parallel poll response for address <a> */


/* Per-card state variables.

   This structure holds the interface card hardware state.  It is replicated for
   each instance of the disc interface.


   Implementation notes:

    1. The FIFO is implemented as a circular queue.  To permit the FIFO to be
       viewed as expected when an EXAMINE command is issued, we maintain
       pointers to the FIFO and TAG register entries, which have the REG_CIRC
       flag, in order to adjust the queue pointers appropriately for FIFO loads
       and unloads.
*/

typedef struct {
    FLIP_FLOP      control;                     /* control flip-flop */
    FLIP_FLOP      flag;                        /* flag flip-flop */
    FLIP_FLOP      flag_buffer;                 /* flag buffer flip-flop */
    FLIP_FLOP      srq;                         /* SRQ flip-flop */
    FLIP_FLOP      edt;                         /* EDT flip-flop */
    FLIP_FLOP      eor;                         /* EOR flip-flop */
    BYTE_SELECTOR  ibp;                         /* input byte pointer selector */
    BYTE_SELECTOR  obp;                         /* output byte pointer selector */

    HP_WORD        cntl_register;               /* control word register */
    HP_WORD        status_register;             /* status word register */
    HP_WORD        input_data_register;         /* input data register */

    HP_WORD        fifo [FIFO_SIZE];            /* FIFO buffer */
    uint32         fifo_count;                  /* FIFO occupancy counter */
    REG            *fifo_reg;                   /* FIFO register pointer */
    REG            *tag_reg;                    /* TAG register pointer */

    BUS_CNTL       bus_cntl;                    /* HP-IB bus control state (ATN, EOI, etc.) */
    double         ifc_timer;                   /* 100 microsecond IFC timer */
    CARD_ID        other;                       /* ID of the cross-connected card for diagnostics */
    } DI_STATE;


/* Disc interface VM global register definitions.

   These definitions should be included before any device-specific registers.


   Implementation notes:

    1. The TMR register is included to ensure that the IFC timer is saved by a
       SAVE command.  SIMH does not provide native floating-pointer register
       examination or deposit, but the SVDATA macro hides the register, which
       otherwise would appear as a byte array.
*/

#define DI_REGS(dev)                                                                              \
/*    Macro   Name    Location                      Radix  Width  Offset    Depth      Flags   */ \
/*    ------  ------  ----------------------------  -----  -----  ------  ---------   -------- */ \
    { ORDATA (CWR,    di [dev].cntl_register,               16)                                }, \
    { ORDATA (SWR,    di [dev].status_register,             16)                                }, \
    { ORDATA (IDR,    di [dev].input_data_register,         16)                                }, \
                                                                                                  \
    { DRDATA (FCNT,   di [dev].fifo_count,                   5),                      PV_LEFT  }, \
    { GBDATA (FIFO,   di [dev].fifo,                  8,    16,     0,    FIFO_SIZE), REG_CIRC }, \
    { GBDATA (FTAG,   di [dev].fifo,                  2,     4,    16,    FIFO_SIZE), REG_CIRC }, \
                                                                                                  \
    { GRDATA (BUSCTL, di [dev].bus_cntl,              2,     8,     8)                         }, \
                                                                                                  \
    { FLDATA (CTL,    di [dev].control,                             0)                         }, \
    { FLDATA (FLG,    di [dev].flag,                                0)                         }, \
    { FLDATA (FBF,    di [dev].flag_buffer,                         0)                         }, \
    { FLDATA (SRQ,    di [dev].srq,                                 0)                         }, \
    { FLDATA (EDT,    di [dev].edt,                                 0)                         }, \
    { FLDATA (EOR,    di [dev].eor,                                 0)                         }, \
                                                                                                  \
    { SVDATA (TMR,    di [dev].ifc_timer)                                                      }, \
                                                                                                  \
    { ORDATA (SC,     dev##_dib.select_code,                 6),                      REG_HRO  }


/* Disc interface VM global modifier definitions.

   These definitions should be included before any device-specific modifiers.
*/

#define DI_MODS(dev,dib)                                                                                          \
/*    Entry Flags          Value  Print String  Match String  Validation       Display           Descriptor    */ \
/*    -------------------  -----  ------------  ------------  ---------------  ----------------  ------------- */ \
    { MTAB_XDV,              0,   NULL,         "HPIB",       &di_set_cable,   NULL,             (void *) &dev }, \
    { MTAB_XDV,              1,   NULL,         "DIAGNOSTIC", &di_set_cable,   NULL,             (void *) &dev }, \
    { MTAB_XDV,              0,   "CABLE",      NULL,         NULL,            &di_show_cable,   (void *) &dev }, \
                                                                                                                  \
    { MTAB_XDV,              0,   "ADDRESS",    "ADDRESS",    &di_set_address, &di_show_address, (void *) &dev }, \
                                                                                                                  \
    { MTAB_XDV,              1,   "SC",         "SC",         &hp_set_dib,     &hp_show_dib,     (void *) &dib }, \
    { MTAB_XDV | MTAB_NMO,  ~1u,  "DEVNO",      "DEVNO",      &hp_set_dib,     &hp_show_dib,     (void *) &dib }


/* Disc interface global bus routine definitions */

typedef t_bool   ACCEPTOR    (CARD_ID card, BUS_CABLE signals);
typedef BUS_DATA RESPONDER   (CARD_ID card, BUS_CNTL  new_signals);
typedef t_bool   INITIALIZER (CARD_ID card);


/* Disc interface global variables */

extern DI_STATE di [];
extern DEBTAB   di_deb [];


/* Disc interface global SCP support routines */

extern IO_INTERFACE   di_interface;
extern IO_INITIALIZER di_initializer;
extern t_stat         di_reset (DEVICE *dptr);


/* Disc interface global SCP routines */

extern t_stat di_set_address (UNIT *uptr, int32 value, char *cptr, void *desc);
extern t_stat di_set_cable   (UNIT *uptr, int32 value, char *cptr, void *desc);

extern t_stat di_show_address (FILE *st, UNIT *uptr, int32 value, void *desc);
extern t_stat di_show_cable   (FILE *st, UNIT *uptr, int32 value, void *desc);


/* Disc interface global bus routines */

extern t_bool di_bus_source  (CARD_ID card, BUS_CABLE signals);
extern void   di_bus_control (CARD_ID card, uint32 address, BUS_CNTL assert, BUS_CNTL deny);
extern t_bool di_bus_ready   (CARD_ID card);
extern void   di_bus_trace   (DEVICE *dptr, BUS_CABLE signals, ...);


/* Amigo disc global bus routines */

extern ACCEPTOR    da_bus_accept;
extern RESPONDER   da_bus_respond;
extern INITIALIZER da_bus_initialize;


/* CS/80 disc/tape global bus routines */

extern ACCEPTOR    dc_bus_accept;
extern RESPONDER   dc_bus_respond;
extern INITIALIZER dc_bus_initialize;


/* Amigo mag tape global bus routines */

/* Not implemented yet.

extern ACCEPTOR  ma_bus_accept;
extern RESPONDER ma_bus_respond;

*/
