/* hp2100_di_da.c: HP 12821A HP-IB Disc Interface simulator for Amigo disc drives

   Copyright (c) 2011-2022, J. David Bryan

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of the author shall not be
   used in advertising or otherwise to promote the sale, use or other dealings
   in this Software without prior written authorization from the author.

   DA           12821A Disc Interface with Amigo disc drives

   27-Mar-22    JDB     Units 1-3 are now disabled by default for lower overhead
   21-Mar-22    JDB     Reset the return buffer index for Amigo Identify
   11-Mar-22    JDB     Updated DIB for "io_initializer" field addition
   24-Jan-22    JDB     Modified "da_boot" for device vs. unit calling sequence
   15-Jul-21    JDB     Improved TRACE_XFER reporting
   30-Jun-21    JDB     Corrected Identify command detection
   16-Jun-21    JDB     Logically separated SIMH unit numbers from bus addresses
   02-Apr-21    JDB     Changed u3-u6 synonyms to title case to avoid conflicts
   07-Feb-21    JDB     Changed CNVARS register macro from BRDATA to SVDATA
   02-Dec-20    JDB     RESET no longer presets the interface
   11-Jul-18    JDB     Revised I/O model, changed global scope of variables to local
   21-Feb-18    JDB     ATTACH -N now creates a full-size disc image
   11-Jul-17    JDB     Renamed "ibl_copy" to "cpu_ibl"
   15-Mar-17    JDB     Changed DEBUG_PRI calls to tprintfs
   09-Mar-17    JDB     Deprecated LOCKED/WRITEENABLED for PROTECT/UNPROTECT
   27-Feb-17    JDB     ibl_copy no longer returns a status code
   17-Jan-17    JDB     Changed to use new byte accessors in hp2100_defs.h
   13-May-16    JDB     Modified for revised SCP API function parameter types
   04-Mar-16    JDB     Name changed to "hp2100_disclib" until HP 3000 integration
   30-Dec-14    JDB     Added S-register parameters to ibl_copy
   24-Dec-14    JDB     Use T_ADDR_FMT with t_addr values for 64-bit compatibility
                        Removed redundant global declarations
   24-Oct-12    JDB     Changed CNTLR_OPCODE to title case to avoid name clash
   07-May-12    JDB     Cancel the intersector delay if an untalk is received
   29-Mar-12    JDB     First release
   04-Nov-11    JDB     Created DA device

   References:
     - HP 13365 Integrated Controller Programming Guide
         (13365-90901, February 1980)
     - HP 7910 Disc Drive Service Manual
         (07910-90903, April 1981)
     - 12745D Disc Controller (13037) to HP-IB Adapter Kit Installation and Service Manual
         (12745-90911, September 1983)
     - HP's 5 1/4-Inch Winchester Disc Drive Service Documentation
         (09134-90032, August 1983)
     - HP 12992 Loader ROMs Installation Manual
         (12992-90001, April 1986)
     - RTE Driver DVA32 Source
         (92084-18708, revision 2540)
     - IEEE Standard Digital Interface for Programmable Instrumentation
         (IEEE-488A-1980, September 1979)


   The HP 7906H, 7920H, and 7925H Integrated Controller Disc (ICD) drives were
   connected via an 12821A disc interface and provided 20MB, 50MB, and 120MB
   capacities.  The drives were identical to the 7906M, 7920M, and 7925M
   Multi-Access Controller (MAC) units but incorporated internal two-card
   controllers in each drive and connected to the CPU interface via the
   Hewlett-Packard Interface Bus (HP-IB), HP's implementation of IEEE-488.  Each
   controller was dedicated to a single drive and operated similarly to the
   12745 Disc Controller to HP-IB Adapter option for the 13037 Disc Controller
   chassis.  The 7906H was introduced in 1980 (there was no 7905H version, as
   the 7905 was obsolete by that time).  Up to four ICD drives could be
   connected to a single 12821A card.  The limitation was imposed by the bus
   loading and the target data transfer rate.

   The ICD command set essentially was the MAC command set modified for
   single-unit operation.  The unit number and CPU hold bit fields in the opcode
   words were unused in the ICD implementation.  The Load TIO Register, Wakeup,
   and Request Syndrome commands were removed, as Load TIO was used with the HP
   3000, Wakeup was used in a multi-CPU environment, and the simpler ICD
   controller did not support ECC.  Controller status values 02B (Unit
   Available) and 27B (Unit Unavailable) were dropped as the controller
   supported only single units, 12B (I/O Program Error) was reused to indicate
   HP-IB protocol errors, 13B (Sync Not Received) was added, and 17B (Possibly
   Correctable Data Error) was removed as error correction was not supported.

   Some minor redefinitions also occurred.  For example, status 14B (End of
   Cylinder) was expanded to include an auto-seek beyond the drive limits, and
   37B (Drive Attention) was restricted to just head unloads from head loads and
   unloads.

   The command set was expanded to include several commands related to HP-IB
   operation.  These were, in large part, adapted from the Amigo disc command
   protocol outlined in the service manual for the HP 9133/34/35 series of
   5-1/4" Winchester drives.  They include the Amigo Identify and Amigo Clear
   sequences, Read and Write Loopback channel tests, and controller Self Test
   commands.


   This simulator implements the Amigo disc protocol.  It calls and is called by
   the 12821A Disc Interface (DI) simulator to send and receive bytes across the
   HP-IB to and from the CPU, and it calls the HP Disc Library to implement the
   controller functions related to disc unit operation (e.g., seek, read, write,
   etc.).  Four units are provided, and any combination of 7906H/20H/25H drives
   may be defined.

   Unfortunately, the primary reference for the ICD controller (the HP 13365
   Integrated Controller Programming Guide) does not indicate parallel poll
   responses for these HP-IB commands.  Therefore, the responses have been
   derived from the sequences in the 7910 and 12745 manuals, although they
   sometimes conflict.

   The drives respond to the following commands.  In the following table, the
   bus addressing state is indicated by U [untalk], L [listen], and T [talk],
   and the secondary and opcode numeric values are in hex.

   The commands that are implemented are:

     Bus  Sec  Op  Operation
     ---  ---  --  ---------------------------------------
      U   MSA  --  Amigo Identify
      L    -   04  Selected Device Clear (primary command)

      L   00   --  Write Data
      L   08   00  Cold Load Read
      L   08   01  Recalibrate
      L   08   02  Seek
      L   08   03  Request Status
      L   08   04  Request Sector Address
      L   08   05  Read
      L   08   06  Read Full Sector
      L   08   07  Verify
      L   08   08  Write
      L   08   09  Write Full Sector
      L   08   0A  Clear
      L   08   0B  Initialize
      L   08   0C  Address Record
      L   08   0E  Read with Offset
      L   08   0F  Set File Mask
      L   08   12  Read without Verify
      L   08   14  Request Logical Disc Address
      L   08   15  End
      L   09   --  Cyclic Redundancy Check
      L   10   --  Amigo Clear
      L   1E   --  Write Loopback
      L   1F   ss  Initiate Self-Test <ss>

      T   00   --  Read Data
      T   08   --  Read Status
      T   09   --  Cyclic Redundancy Check
      T   10   --  Device Specified Jump
      T   1E   --  Read Loopback
      T   1F   --  Return Self-Test Result


   The HP-IB bus transaction sequence for each of the commands is listed below:

     Identify
     --------

       ATN  UNT     Untalk
       ATN  MSA     My secondary address
            DAB     ID data byte #1
       EOI  DAB     ID data byte #2
       ATN  OTA     Talk 30


     Selected Device Clear
     ---------------------

       ATN  MLA     My listen address
       ATN  PCG     Primary command 04H
            ppe     Parallel poll enabled when clear completes
       ATN  UNL     Unlisten


     Amigo Clear
     -----------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 10H (Amigo Clear)
            ppd     Parallel poll disabled
       EOI  DAB     Unused data byte
       ATN  SDC     Selected device clear
            ppe     Parallel poll enabled when clear completes
       ATN  UNL     Unlisten


     CRC
     ---

       ATN  MTA     My talk address
       ATN  SCG     Secondary command 09H
            ppd     Parallel poll disabled
            DAB     Data byte #1
            ...
       EOI  DAB     Data byte #n
            ppe     Parallel poll enabled
       ATN  UNT     Untalk

     or

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 09H
            ppd     Parallel poll disabled
            DAB     Data byte #1
            ...
       EOI  DAB     Data byte #n
            ppe     Parallel poll enabled
       ATN  UNL     Unlisten


     Device Specified Jump
     ---------------------

       ATN  MTA     My talk address
       ATN  SCG     Secondary command 10H
            ppd     Parallel poll disabled
       EOI  DAB     DSJ data byte
       ATN  UNT     Untalk


     Initiate Self-Test and Return Self-Test Result
     ----------------------------------------------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 1FH
            ppd     Parallel poll disabled
       EOI  DAB     Self-test number
            ppe     Parallel poll enabled
       ATN  UNL     Unlisten

       ATN  MTA     My talk address
       ATN  SCG     Secondary command 1FH
            ppd     Parallel poll disabled
       EOI  DAB     Result data byte
            ppe     Parallel poll enabled
       ATN  UNT     Untalk


     Write Loopback and Read Loopback
     --------------------------------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 1EH
            ppd     Parallel poll disabled
            DAB     Loopback data byte #1
            ...
       EOI  DAB     Loopback data byte #256
            ppe     Parallel poll enabled
       ATN  UNL     Unlisten

       ATN  MTA     My talk address
       ATN  SCG     Secondary command 1EH
            ppd     Parallel poll disabled
            DAB     Loopback data byte #1
            ...
       EOI  DAB     Loopback data byte #16
            ppe     Parallel poll enabled
       ATN  UNT     Untalk


     Recalibrate and Seek
     --------------------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 08H
            ppd     Parallel poll disabled
            DAB     Opcode 01H, 02H
            ...     (one to five
       EOI  DAB        parameter bytes)
       ATN  UNL     Unlisten
            ...
            ppe     Parallel poll enabled when seek completes


     Clear, Address Record, and Set File Mask
     ----------------------------------------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 08H
            ppd     Parallel poll disabled
            DAB     Opcode 0AH, 0CH, 0FH
            ...     (one to five
       EOI  DAB        parameter bytes)
            ppe     Parallel poll enabled
       ATN  UNL     Unlisten


     End
     ---

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 08H
            ppd     Parallel poll disabled
            DAB     Opcode 15H
       EOI  DAB     Unused data byte
       ATN  UNL     Unlisten


     Request Status, Request Sector Address, and Request Logical Disc Address
     ------------------------------------------------------------------------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 08H
            ppd     Parallel poll disabled
            DAB     Opcode 03H, 04H, 14H
       EOI  DAB     Unused data byte
       ATN  UNL     Unlisten

       ATN  MTA     My talk address
       ATN  SCG     Secondary command 08H
            DAB     Status byte #1
            ...     (two to four
       EOI  DAB        status bytes)
            ppe     Parallel poll enabled
       ATN  UNT     Untalk


   Cold Load Read, Read, Read Full Sector,
   Verify, Read with Offset, and Read without Verify
   -------------------------------------------------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 08H
            ppd     Parallel poll disabled
            DAB     Opcode 00H, 05H, 06H, 07H, 0EH, 12H
       EOI  DAB     Unused data byte
       ATN  UNL     Unlisten

       ATN  MTA     My talk address
       ATN  SCG     Secondary command 00H
            DAB     Read data byte #1
            ...
            DAB     Read data byte #n
       ATN  UNT     Untalk
            ...
            ppe     Parallel poll enabled when sector ends


   Write, Write Full Sector, and Initialize
   ----------------------------------------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 08H
            ppd     Parallel poll disabled
            DAB     Opcode 08H, 09H, 0BH
       EOI  DAB     Unused data byte
       ATN  UNL     Unlisten

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 00H
            DAB     Write data byte #1
            ...
       EOI  DAB     Write data byte #n
            ppe     Parallel poll enabled
       ATN  UNL     Unlisten


   Disc image files are implemented as a bare stream of data bytes, with each
   16-bit word written as two bytes in little-endian format, regardless of the
   host platform convention.  Disc sector address headers and CRC/ECC trailers
   are not implemented.  The image file format is compatible with the format
   used by the 79xx MAC simulator (DS device) but NOT with the format used by
   HPDrive, a disc/tape emulation program written by Ansgar Kueckes.  HPDrive
   interchange is possible if each pair of bytes is swapped in the file to
   produce big-endian 16-bit words.  The GNU Coreutils "dd" program will do
   this, for example.


   Implementation notes:

    1. The 12745 does not alter the parallel poll response for the
       Device-Specified Jump command.

    2. The 7910 does not perform a parallel poll response enable and disable
       between the Initiate Self-Test and Return Self-Test Result commands.

    3. The 12745 does not disable the parallel poll response for the Read
       Loopback command.
*/



#include "hp2100_defs.h"
#include "hp2100_io.h"
#include "hp2100_di.h"
#include "hp2100_disclib.h"



/* Program constants */

#define DA_UNITS            4                   /* number of addressable disc units */
#define DEVICE_COUNT        8                   /* count of bus device addresses */


/* Unit flags.

   UNIT_V_UF + 15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
   DL_V_UF   +  9   8   7   6   5   4   3   2   1   0
              +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
              | - | - | - | - | - | - | - |  address  |      (reserved)       |
              +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   "address" is the HP-IB address of the simulated disc unit (0-7).  The lower
   flag bits are reserved for use by the HP disc library.
*/

#define UNIT_ADDRESS_SHIFT  (DL_V_UF + 0)

#define UNIT_ADDRESS_MASK   007u

#define UNIT_ADDRESS        (UNIT_ADDRESS_MASK << UNIT_ADDRESS_SHIFT)

#define ADDRESS(f)          ((f) >> UNIT_ADDRESS_SHIFT & UNIT_ADDRESS_MASK)

#define TO_ADDRESS(w)       ((w) << UNIT_ADDRESS_SHIFT & UNIT_ADDRESS)


/* Interface states */

typedef enum {
    idle = 0,                                   /* idle = default for reset */
    opcode_wait,                                /* waiting for opcode reception */
    parameter_wait,                             /* waiting for parameter reception */
    read_wait,                                  /* waiting for send read data secondary */
    write_wait,                                 /* waiting for receive write data secondary */
    status_wait,                                /* waiting for send status secondary */
    command_exec,                               /* executing an interface command */
    command_wait,                               /* waiting for command completion */
    read_xfer,                                  /* sending read data or status */
    write_xfer,                                 /* receiving write data */
    error_source,                               /* sending bytes for error recovery */
    error_sink                                  /* receiving bytes for error recovery */
    } IF_STATE;


/* Interface state names */

static const char * const if_state_name [] = {  /* state names, indexed by IF_STATE */
    "idle",
    "opcode wait",
    "parameter wait",
    "read wait",
    "write wait",
    "status wait",
    "command execution",
    "command wait",
    "read transfer",
    "write transfer",
    "error source",
    "error sink"
    };


/* Next interface state after command recognition */

static const IF_STATE next_state [] = {
    read_wait,                                  /* cold load read */
    command_exec,                               /* recalibrate */
    command_exec,                               /* seek */
    status_wait,                                /* request status */
    status_wait,                                /* request sector address */
    read_wait,                                  /* read */
    read_wait,                                  /* read full sector */
    command_exec,                               /* verify */
    write_wait,                                 /* write */
    write_wait,                                 /* write full sector */
    command_exec,                               /* clear */
    write_wait,                                 /* initialize */
    command_exec,                               /* address record */
    idle,                                       /* request syndrome */
    read_wait,                                  /* read with offset */
    command_exec,                               /* set file mask */
    idle,                                       /* invalid */
    idle,                                       /* invalid */
    read_wait,                                  /* read without verify */
    idle,                                       /* load TIO register */
    status_wait,                                /* request disc address */
    command_exec,                               /* end */
    idle                                        /* wakeup */
    };


/* Interface commands */

typedef enum {
    invalid = 0,                                /* invalid = default for reset */
    disc_command,                               /* MLA 08 */
    crc_listen,                                 /* MLA 09 */
    amigo_clear,                                /* MLA 10 */
    write_loopback,                             /* MLA 1E */
    initiate_self_test,                         /* MLA 1F */
    crc_talk,                                   /* MTA 09 */
    device_specified_jump,                      /* MTA 10 */
    read_loopback,                              /* MTA 1E */
    return_self_test_result,                    /* MTA 1F */
    amigo_identify                              /* UNT MSA */
    } IF_COMMAND;

/* Interface command names */

static const char * const if_command_name [] = {
    "invalid",
    "disc command",
    "CRC listen",
    "Amigo clear",
    "write loopback",
    "initiate self-test",
    "CRC talk",
    "device specified jump",
    "read loopback",
    "return self-test result",
    "Amigo identify"
    };



/* Amigo disc state variables */

typedef struct {                                /* channel state */
    uint32    present;                          /*   bus address bitmap of devices present on the bus */
    uint32    listeners;                        /*   bus address bitmap of the listeners */
    uint32    talker;                           /*   bus address of the talker */
    t_bool    poll_active;                      /*   TRUE if a parallel poll is in progress */
    BUS_DATA  poll_response;                    /*   set of parallel poll responses */
    BUS_DATA  command;                          /*   last channel command */
    } CHANNEL_STATE;

static CHANNEL_STATE channel;                   /* channel state for all devices */

static uint16 buffer [DL_BUFSIZE];              /* command/status/sector buffer */

static uint8      if_dsj     [DA_UNITS];        /* ICD controller DSJ values */
static IF_STATE   if_state   [DA_UNITS];        /* ICD controller state */
static IF_COMMAND if_command [DA_UNITS];        /* ICD controller command */

static CNTLR_VARS icd_cntlr [DA_UNITS] =        /* ICD controllers: */
    { { CNTLR_INIT (ICD, buffer, NULL) },       /*   unit 0 controller */
      { CNTLR_INIT (ICD, buffer, NULL) },       /*   unit 1 controller */
      { CNTLR_INIT (ICD, buffer, NULL) },       /*   unit 2 controller */
      { CNTLR_INIT (ICD, buffer, NULL) } };     /*   unit 3 controller */


/* Amigo disc local SCP support routine declarations */

static t_stat da_service (UNIT   *uptr);
static t_stat da_reset   (DEVICE *dptr);
static t_stat da_attach  (UNIT   *uptr,  char *cptr);
static t_stat da_detach  (UNIT   *uptr);
static t_stat da_boot    (int32  unitno, DEVICE *dptr);

static t_stat da_load_unload (UNIT *uptr, int32 value, char *cptr, void *desc);
static t_stat da_set_address (UNIT *uptr, int32 value, char *cptr, void *desc);

static t_stat da_show_address (FILE *st, UNIT *uptr, int32 value, void *desc);


/* Amigo disc local utility routine declarations */

static void   process_secondary (uint32 address, BUS_CABLE bus);
static void   process_data      (uint32 address, BUS_CABLE bus);
static t_bool start_command     (uint32 unit);
static void   abort_command     (uint32 unit, CNTLR_STATUS status, IF_STATE state);
static void   complete_read     (uint32 unit);
static void   complete_write    (uint32 unit);
static void   complete_abort    (uint32 unit);
static uint8  get_buffer_byte   (CVPTR  cvptr);
static void   put_buffer_byte   (CVPTR  cvptr, uint8 data);
static t_stat activate_unit     (UNIT   *uptr);
static void   set_response      (uint32 address, FLIP_FLOP response);


/* Amigo disc SCP data declarations */

DEVICE da_dev;


/* Device information block */

static DIB da_dib = {
    di_interface,                               /* the device's I/O interface function pointer */
    di_initializer,                             /* the execution initialization function pointer */
    NULL,                                       /* the power state function pointer */
    DI_DA,                                      /* the device's select code (02-77) */
    da,                                         /* the card index */
    "12821A Disc Interface",                    /* the card description */
    "12992H 7906H/7920H/7925H/9895 Disc Loader" /* the ROM description */
    };


/* Unit list.

   SIMH units are enabled and assigned to Amigo units as needed.  For
   efficiency, only one unit is initially enabled.
*/

#define UNIT_FLAGS  (UNIT_FIX | UNIT_ATTABLE | UNIT_ROABLE | UNIT_DISABLE | UNIT_UNLOAD)

static UNIT da_unit [] = {                      /* unit array, indexed by SIMH unit number */
/*           Event Routine  Unit Flags                                           Capacity  */
/*           -------------  ---------------------------------------------------  --------- */
    { UDATA (&da_service,   MODEL_7906 | TO_ADDRESS (0) | UNIT_FLAGS,            D7906_WORDS) },
    { UDATA (&da_service,   MODEL_7906 | TO_ADDRESS (1) | UNIT_FLAGS | UNIT_DIS, D7906_WORDS) },
    { UDATA (&da_service,   MODEL_7906 | TO_ADDRESS (2) | UNIT_FLAGS | UNIT_DIS, D7906_WORDS) },
    { UDATA (&da_service,   MODEL_7906 | TO_ADDRESS (3) | UNIT_FLAGS | UNIT_DIS, D7906_WORDS) }
    };


/* Register list.

   Most of the registers are provided by the 12821A Disc Interface card via the
   DI_REGS macro.  Amigo hardware devices do not have externally accessible
   registers, but we provide a few "pseudo-registers" to provide a view into the
   device controller state.


   Implementation notes:

    1. The CHANNEL and CONTROL registers are included to ensure that the state
       variables arrays are saved by a SAVE command.
*/

static REG da_reg [] = {
/*    Macro   Name     Location    Field     Radix  Width  Offset  Depth       Flags            */
/*    ------  -------- ----------  --------  -----  -----  ------  ----------  ---------------- */
      DI_REGS (da),

    { BRDATA (BUFFER,  buffer,                 8,    16,           DL_BUFSIZE)                  },

    { BRDATA (DSJ,     if_dsj,                10,     2,           DA_UNITS)                    },
    { BRDATA (ISTATE,  if_state,              10,     4,           DA_UNITS)                    },
    { BRDATA (ICMD,    if_command,            10,     4,           DA_UNITS)                    },

    { SRDATA (CSTATE,  icd_cntlr,  state,     10,     2,     0,    DA_UNITS),  REG_RO           },
    { SRDATA (OPCODE,  icd_cntlr,  opcode,     8,     6,     0,    DA_UNITS),  REG_RO           },
    { SRDATA (STATUS,  icd_cntlr,  status,     8,     6,     0,    DA_UNITS),  REG_RO           },
    { SRDATA (CYL,     icd_cntlr,  cylinder,  10,    16,     0,    DA_UNITS),  REG_RO | PV_LEFT },
    { SRDATA (HEAD,    icd_cntlr,  head,      10,     6,     0,    DA_UNITS),  REG_RO | PV_LEFT },
    { SRDATA (SECTOR,  icd_cntlr,  sector,    10,     8,     0,    DA_UNITS),  REG_RO | PV_LEFT },

    { SVDATA (CHANNEL, channel)                                                                 },
    { SVDATA (CONTROL, icd_cntlr)                                                               },

    { NULL }
    };


/* Modifier list */

static MTAB da_mod [] = {
    DI_MODS (da_dev, da_dib),

/*    Mask Value    Match Value   Print String       Match String     Validation        Display  Descriptor */
/*    ------------  ------------  -----------------  ---------------  ----------------  -------  ---------- */
    { UNIT_UNLOAD,  UNIT_UNLOAD,  "heads unloaded",  "UNLOADED",      &da_load_unload,  NULL,    NULL       },
    { UNIT_UNLOAD,  0,            "heads loaded",    "LOADED",        &da_load_unload,  NULL,    NULL       },

    { UNIT_WLK,     UNIT_WLK,     "protected",       "PROTECT",       NULL,             NULL,    NULL       },
    { UNIT_WLK,     0,            "unprotected",     "UNPROTECT",     NULL,             NULL,    NULL       },

    { UNIT_WLK,     UNIT_WLK,     NULL,              "LOCKED",        NULL,             NULL,    NULL       },
    { UNIT_WLK,     0,            NULL,              "WRITEENABLED",  NULL,             NULL,    NULL       },

    { UNIT_FMT,     UNIT_FMT,     "format enabled",  "FORMAT",        NULL,             NULL,    NULL       },
    { UNIT_FMT,     0,            "format disabled", "NOFORMAT",      NULL,             NULL,    NULL       },

    { UNIT_MODEL,   MODEL_7906,   "7906H",           "7906H",         &dl_set_model,    NULL,    NULL       },
    { UNIT_MODEL,   MODEL_7920,   "7920H",           "7920H",         &dl_set_model,    NULL,    NULL       },
    { UNIT_MODEL,   MODEL_7925,   "7925H",           "7925H",         &dl_set_model,    NULL,    NULL       },

/*    Entry Flags  Value  Print String  Match String  Validation        Display            Descriptor */
/*    -----------  -----  ------------  ------------  ----------------  -----------------  ---------- */
    { MTAB_XUN,      0,   "BUS",        "BUS",        &da_set_address,  &da_show_address,  NULL       },

    { 0 }
    };


/* Device descriptor */

DEVICE da_dev = {
    "DA",                                       /* device name */
    da_unit,                                    /* unit array */
    da_reg,                                     /* register array */
    da_mod,                                     /* modifier array */
    DA_UNITS,                                   /* number of units */
    10,                                         /* address radix */
    26,                                         /* address width */
    1,                                          /* address increment */
    8,                                          /* data radix */
    16,                                         /* data width */
    NULL,                                       /* examine routine */
    NULL,                                       /* deposit routine */
    &da_reset,                                  /* reset routine */
    &da_boot,                                   /* boot routine */
    &da_attach,                                 /* attach routine */
    &da_detach,                                 /* detach routine */
    &da_dib,                                    /* device information block */
    DEV_DISABLE | DEV_DEBUG | TO_DEV_ADDR (0),  /* device flags */
    0,                                          /* debug control flags */
    di_deb,                                     /* debug flag name table */
    NULL,                                       /* memory size change routine */
    NULL                                        /* logical device name */
    };



/* Amigo disc global VM routines */


/* Service an Amigo disc drive I/O event.

   This service routine is called to execute commands and control the transfer
   of data to and from the HP-IB card.  The actions to be taken depend on the
   current state of the ICD interface.  The possibilities are:

    1. A command is pending on the interface.  This occurs only when a command
       is received while a Seek or Recalibrate command is in progress.

    2. A command is executing.

    3. Data is being sent or received over the HP-IB during command execution.

    4. Dummy bytes are being sent or received over the HP-IB due to a command
       error.

   Entry to the the service routine in any other interface state or to process a
   command not allowed in a valid state will return an Internal Error to cause a
   simulator stop.  Exit from the routine will be either in one of the above
   states, or in the idle state if the operation is complete.

   The specific actions taken for the various interface states are as follows:

     command_wait
     ------------
     We are entered in this state only if a unit that was busy (still seeking)
     was addressed to listen or talk.  The card has been held off by asserting
     NRFD after receiving MLA or MTA.  Upon entry, we complete the seek and then
     release the interface by denying NRFD to allow the remainder of the command
     sequence to be received from the card.

     command_exec
     ------------
     We are entered in this state to initiate, continue, or complete a command.
     The command may be a disc command, such as Seek or Read, or an interface
     command, such as Amigo Identify or Device-Specified Jump.

     Disc commands call the disc library service routine to perform all of the
     common controller actions.  Any ICD-specific actions needed, such as
     setting the DSJ value, are performed after the call.

     Certain disc commands require multiple execution phases.  For example, the
     Read command has a start phase that reads data from the disc image file
     into the sector buffer, a data phase that transfers bytes from the buffer
     to the card, and an end phase that schedules the intersector gap time and
     resets to the start phase.  Data phase transfers are performed in the
     read_xfer or write_xfer interface states.

     The results of the disc library service are inferred by the controller
     state.  If the controller is busy, then the command continues in a new
     phase.  Otherwise, the command either has completed normally or has
     terminated with an error.  If an error has occurred during a disc command
     that transfers data, DSJ is set to 1, and the interface state is changed to
     source or sink dummy bytes to complete the command sequence.

     Interface commands may either complete immediately (e.g., Amigo Clear) or
     transfer data (e.g., DSJ).

     read_xfer
     ---------
     Commands that send data to the CPU enter the service routine to source a
     byte to the bus.  Bytes are transferred only when ATN and NRFD are denied;
     if they are not, we simply exit, as we will be rescheduled when the lines
     are dropped.  Otherwise, we get a byte from the sector buffer and send it
     to the card.  If the card has stopped listening, or the buffer is now
     empty, then we terminate the transfer and move to the end phase of the
     command.  Otherwise, we reschedule the next data phase byte transfer.

     Disc and interface commands are handled separately, as EOI is always
     asserted on the last byte of an interface command transfer and never on a
     (good) disc command transfer.

     write_xfer
     ----------
     Commands that receive data from the CPU enter the service routine to
     determine whether or not to continue the transfer.  Our bus accept routine
     has already stored the received byte in the sector buffer and has asserted
     NRFD to hold off the card.  If the buffer is now full, or the byte was
     tagged with EOI, then we terminate the transfer and move to the end phase
     of the command.  Otherwise, we deny NRFD and exit; we will be rescheduled
     when the next byte arrives.

     error_source
     ------------
     If an error occurred during the data transfer phase of a read or status
     command, a dummy byte tagged with EOI is sourced to the bus.  This allows
     the OS driver for the card to terminate the command and request the
     controller's status.

     error_sink
     ----------
     If an error occurred during the data transfer phase of a write command,
     dummy bytes are sunk from the bus until EOI is seen or the card is
     unaddressed.  This allows the OS driver to complete the command as expected
     and then determine the cause of the failure by requesting the controller's
     status.


   Implementation notes:

    1. The disc library sets the controller state to idle for a normal End,
       Seek, or Recalibrate command and to wait for all other commands that end
       normally.  So we determine command completion by checking if the
       controller is not busy, rather than checking if the controller is idle.

       Drive Attention status is the normal result of the completion of a Seek
       or Recalibrate command.  Normal Completion status is the normal result of
       all other commands.

    2. The disc library returns the buffer length in words.  We double the
       return value to count bytes.

    3. Some commands, such as DSJ, could be completed in the bus accept routine.
       They are serviced here instead to avoid presenting a zero execution time
       to the CPU.

    4. The Amigo command set does not provide the disc with the number of bytes
       that will be read, and the unit expects to be untalked when the read is
       to terminate.  The RTE ICD bootstrap extension does not do this.
       Instead, it resets the card via CLC 0,C to terminate the Cold Load Read
       that was started by the ICD boot loader ROM.

       In hardware, if the LSTN control bit is cleared, e.g., by CRS,
       transmission stops because the card denies NDAC and NRFD (the HP-IB
       handshake requires NDAC and NRFD to be asserted to start the handshake
       sequence; TACS * SDYS * ~NDAC * ~NRFD is an error condition).  In
       simulation, we handle this by terminating a read transfer if the card
       stops accepting.  If we did not, then the disc would continue to source
       bytes to the bus, overflowing the card FIFO (a FIFO full condition cannot
       assert NRFD if the LSTN control bit is clear).
*/

static t_stat da_service (UNIT *uptr)
{
const int32 unit = uptr - da_unit;              /* the disc unit number */
const CVPTR cvptr = &icd_cntlr [unit];          /* a pointer to the controller */
CNTLR_CLASS command_class;
uint32      bus;
IF_STATE    last_state = if_state [unit];
t_stat      result = SCPE_OK;
t_bool      release_interface = FALSE;

tprintf (da_dev, TRACE_SERV, "Unit %u service entered in %s state\n",
         unit, if_state_name [last_state]);

switch (if_state [unit]) {                              /* dispatch the interface state */

    case command_wait:                                  /* command is waiting */
        release_interface = TRUE;                       /* release the interface at then end if it's idle */

    /* fall through into the command_exec handler to process the current command */

    case command_exec:                                  /* command is executing */
        switch (if_command [unit]) {                    /* dispatch the interface command */

            case disc_command:                              /* execute a disc command */
                result = dl_service_drive (cvptr, uptr);    /* service the disc unit */

                if (cvptr->opcode == Clear)             /* is this a Clear command? */
                    if_dsj [unit] = 2;                  /* indicate that the self test is complete */

                if (cvptr->state != cntlr_busy) {       /* has the controller stopped? */
                    if_state [unit] = idle;             /* idle the interface */

                    if (cvptr->status == normal_completion ||   /* do we have normal completion */
                        cvptr->status == drive_attention)       /*   or drive attention? */
                        break;                                  /* we're done */

                    else {                              /* if the status is abnormal */
                        if_dsj [unit] = 1;              /*   an error has occurred */

                        command_class = dl_classify (*cvptr);   /* classify the command */

                        if (command_class == class_write) { /* did a write command fail? */
                            if_state [unit] = error_sink;   /* sink the remaining bytes */
                            uptr->wait = cvptr->cmd_time;   /* activate to complete processing */
                            }

                        else if (command_class != class_control) {  /* did a read or status command fail? */
                            if_state [unit] = error_source;         /* source an error byte */
                            uptr->wait = cvptr->cmd_time;           /* activate to complete processing */
                            }
                        }
                    }

                else if (uptr->Phase == data_phase) {   /* are we starting the data phase? */
                    cvptr->length = cvptr->length * 2;  /* convert the buffer length to bytes */

                    if (dl_classify (*cvptr) == class_write)    /* is this a write command? */
                        if_state [unit] = write_xfer;           /* set for a write data transfer */
                    else                                        /* it is a read or status command */
                        if_state [unit] = read_xfer;            /* set for a read data transfer */
                    }

                break;


            case amigo_identify:                        /* Amigo Identify */
                buffer [0] = 0x0003;                    /* store the response in the buffer */
                cvptr->length = 2;                      /* return two bytes */

                if_state [unit] = read_xfer;            /* we are ready to transfer the data */
                uptr->wait = cvptr->cmd_time;           /* schedule the transfer */

                tprintf (da_dev, TRACE_INCO, "Unit %d Amigo identify response %04XH\n",
                         unit, buffer [0]);
                break;


            case initiate_self_test:                    /* Initiate a self test */
                sim_cancel (&da_unit [unit]);           /* cancel any operation in progress */
                dl_clear_controller (cvptr,             /* hard-clear the controller */
                                     &da_unit [unit],
                                     hard_clear);
                if_dsj [unit] = 2;                      /* set DSJ for self test completion */
                if_state [unit] = idle;                 /* the command is complete */
                set_response (unit, SET);               /*   with PPR enabled */
                break;


            case amigo_clear:                           /* Amigo clear */
                dl_idle_controller (cvptr);             /* idle the controller */
                if_dsj [unit] = 0;                      /* clear the DSJ value */
                if_state [unit] = idle;                 /* the command is complete */
                set_response (unit, SET);               /*   with PPR enabled */
                break;


            default:                                    /* no other commands are executed */
                result = SCPE_IERR;                     /* signal an internal error */
                break;
            }                                           /* end of command dispatch */
        break;


    case error_source:                                  /* send data after an error */
        if (di_bus_ready (da)) {                        /* is the card ready for data? */
            di_bus_source (da, 0 | Bus_EOI);            /*   and send a dummy byte with EOI to the card */
            if_state [unit] = idle;                     /* the command is complete */
            }
        break;


    case read_xfer:                                     /* send read data */
        if (di_bus_ready (da))                          /* is the card ready for data? */
            switch (if_command [unit]) {                /* dispatch the interface command */

                case disc_command:                      /* disc read or status commands */
                    bus = get_buffer_byte (cvptr);      /* get the next byte from the buffer */

                    if (di_bus_source (da, bus) == FALSE)   /* send the byte to the card; is it listening? */
                        cvptr->eod = SET;                   /* no, so terminate the read */

                    if (cvptr->length == 0 || cvptr->eod == SET) {  /* is the data phase complete? */
                        uptr->Phase = end_phase;                    /* set the end phase */

                        if (cvptr->opcode == Request_Status)    /* is it a Request Status command? */
                            if_dsj [unit] = 0;                  /* clear the DSJ value */

                        if_state [unit] = command_exec; /* set to execute the command */
                        uptr->wait = cvptr->cmd_time;   /*   and reschedule the service */
                        }

                    else                                /* the data phase continues */
                        uptr->wait = cvptr->data_time;  /* reschedule the next transfer */

                    break;


                case amigo_identify:
                case read_loopback:
                case return_self_test_result:
                    bus = get_buffer_byte (cvptr);      /* get the next byte from the buffer */

                    if (cvptr->length == 0)             /* is the transfer complete? */
                        bus |= Bus_EOI;                 /* set EOI */

                    if (di_bus_source (da, bus)         /* send the byte to the card; is it listening? */
                      && cvptr->length > 0)             /*   and is there more to transfer? */
                        uptr->wait = cvptr->data_time;  /* reschedule the next transfer */

                    else {                              /* the transfer is complete */
                        if_state [unit] = idle;         /* the command is complete */
                        set_response (unit, SET);       /* enable the PPR */

                        if (if_command [unit] == amigo_identify)    /* if this was an Identify */
                            channel.talker = BUS_UNADDRESS;         /*   then unaddress the device */
                        }
                    break;


                case device_specified_jump:
                    di_bus_source (da, if_dsj [unit] | Bus_EOI);    /* send the DSJ value to the card */
                    if_state [unit] = idle;                         /* the command is complete */
                    break;


                case crc_talk:
                    di_bus_source (da, 0 | Bus_EOI);    /* send dummy bytes */
                    break;                              /*   until the card untalks */


                default:                                /* no other commands send data */
                    result = SCPE_IERR;                 /* signal an internal error */
                    break;
                }                                       /* end of read data transfer dispatch */
        break;


    case error_sink:                                    /* absorb data after an error */
        cvptr->index = 0;                               /* absorb data until EOI asserts */

        if (cvptr->eod == SET)                          /* is the transfer complete? */
            if_state [unit] = idle;                     /* the command is complete */

        di_bus_control (da, unit, 0, Bus_NRFD);         /* deny NRFD to allow the card to resume */
        break;


    case write_xfer:                                    /* receive write data */
        switch (if_command [unit]) {                    /* dispatch the interface command */

            case disc_command:                                  /* disc write commands */
                if (cvptr->length == 0 || cvptr->eod == SET) {  /* is the data phase complete? */
                    uptr->Phase = end_phase;                    /* set the end phase */

                    if_state [unit] = command_exec;     /* set to execute the command */
                    uptr->wait = cvptr->cmd_time;       /*   and schedule the service */

                    if (cvptr->eod == CLEAR)            /* is the transfer continuing? */
                        break;                          /* do not deny NRFD until next service! */
                    }

                di_bus_control (da, unit, 0, Bus_NRFD); /* deny NRFD to allow the card to resume */
                break;


            case write_loopback:
                if (cvptr->eod == SET) {                /* is the transfer complete? */
                    cvptr->length = 16 - cvptr->length; /* set the count of bytes transferred */
                    if_state [unit] = idle;             /* the command is complete */
                    }

                di_bus_control (da, unit, 0, Bus_NRFD); /* deny NRFD to allow the card to resume */
                break;


            default:                                    /* no other commands receive data */
                result = SCPE_IERR;                     /* signal an internal error */
                break;
            }                                           /* end of write data transfer dispatch */
        break;


    default:                                            /* no other states schedule service */
        result = SCPE_IERR;                             /* signal an internal error */
        break;
    }                                                   /* end of interface state dispatch */


if (uptr->wait)                                         /* is service requested? */
    activate_unit (uptr);                               /* schedule the next event */

if (result == SCPE_IERR)                                /* did an internal error occur? */
    if (if_state [unit] == command_exec
      && if_command [unit] == disc_command)
        tprintf (da_dev, TRACE_INCO, "Unit %d %s command %s phase service not handled\n",
                 unit,
                 dl_opcode_name (ICD, (CNTLR_OPCODE) uptr->Op),
                 dl_phase_name ((CNTLR_PHASE) uptr->Phase));
    else
        tprintf (da_dev, TRACE_INCO, "Unit %d %s state %s service not handled\n",
                 unit,
                 if_command_name [if_command [unit]],
                 if_state_name [if_state [unit]]);

if (if_state [unit] == idle) {                          /* is the command now complete? */
    if (if_command [unit] == disc_command) {            /* did a disc command complete? */
        if (cvptr->opcode != End)                       /* yes; if the command was not End, */
            set_response (unit, SET);                   /*   then enable PPR */

        tprintf (da_dev, TRACE_INCO, "Unit %d %s disc command completed with DSJ %u\n",
                 unit, dl_opcode_name (ICD, cvptr->opcode), if_dsj [unit]);
        }

    else                                                /* an interface command completed */
        tprintf (da_dev, TRACE_INCO, "Unit %d %s command completed with DSJ %u\n",
                 unit, if_command_name [if_command [unit]], if_dsj [unit]);

    if (release_interface)                              /* if the next command is already pending */
        di_bus_control (da, unit, 0, Bus_NRFD);         /*   deny NRFD to allow the card to resume */
    }

if (if_state [unit] != last_state)
    tprintf (da_dev, TRACE_STATE, "Unit %u transitioned from %s state to %s state\n",
             unit, if_state_name [last_state], if_state_name [if_state [unit]]);

return result;                                          /* return the result of the service */
}


/* Device reset routine.

   This routine is called for a RESET or RESET DA command.  With the addition of
   the -P switch, it is the simulation equivalent of applying or cycling power
   on the set of bus-connected devices.  Without the switch, only the interface
   card is reset; this action has no direct hardware analog.

   In hardware, a self-test is performed by the controller at power-on.  When
   the self-test completes, the controller sets DSJ = 2 and enables the parallel
   poll response.

   A front panel PRESET or programmed CRS has no direct effect on the controller
   or drive.  However, the card reacts to CRS by clearing its talker and
   listener states, so an in-progress read or status command will abort when the
   next byte sourced to the bus finds no acceptors.
*/

static t_stat da_reset (DEVICE *dptr)
{
uint32 unit;
t_stat status;

status = di_reset (dptr);                               /* reset the card */

if (status == SCPE_OK && (sim_switches & SWMASK ('P'))) /* is the card OK and is this a power-on reset? */
    for (unit = 0; unit < dptr->numunits; unit++) {     /* loop through the units */
        sim_cancel (dptr->units + unit);                /* cancel any current activation */
        dptr->units [unit].Cyl = 0;                     /* reset the head position */
        dptr->units [unit].pos = 0;                     /*   to cylinder 0 */

        dl_clear_controller (&icd_cntlr [unit],         /* hard-clear the controller */
                             dptr->units + unit,
                             hard_clear);

        if_state [unit] = idle;                         /* reset the interface state */
        if_command [unit] = invalid;                    /* reset the interface command */

        if_dsj [unit] = 2;                              /* set the DSJ for power up complete */
        }

return status;
}


/* Attach a disc image file to a unit.

   The file named by the "cptr" parameter is attached to the unit specified by
   the "uptr" parameter.  This is the simulation equivalent of inserting a disc
   pack into the disc drive and loading the disc heads.  While a unit is
   unattached, it reports Not Ready error status.  The simulator considers an
   enabled unit to be connected to the bus and a disabled unit to be
   disconnected.  An attached unit is ready if the heads are loaded and not
   ready if not.

   If a new file is specified, the file is initialized to its capacity by
   writing a zero to the last byte in the file.


   Implementation notes:

    1. The C standard says, "A binary stream need not meaningfully support fseek
       calls with a whence value of SEEK_END," so instead we determine the
       offset from the start of the file to the last byte and seek there.
*/

static t_stat da_attach (UNIT *uptr, char *cptr)
{
t_stat      result;
t_addr      offset;
const uint8 zero = 0;
const int32 unit = uptr - da_unit;                      /* the number of the attaching unit */

result = dl_attach (&icd_cntlr [unit], uptr, cptr);     /* attach the drive */

if (result == SCPE_OK)                                  /* if the attach was successful */
    if (sim_switches & SWMASK ('N')) {                  /*   then if this is a new disc image */
        offset = (t_addr)                               /*     then determine the offset of */
          (uptr->capac * sizeof (int16) - sizeof zero); /*       the last byte in a full-sized file */

        if (sim_fseek (uptr->fileref, offset, SEEK_SET) != 0    /* seek to the last byte */
          || fwrite (&zero, sizeof zero, 1, uptr->fileref) == 0 /*   and write a zero to fill */
          || fflush (uptr->fileref) != 0)                       /*     the file to its capacity */
            clearerr (uptr->fileref);                           /* clear and ignore any errors */
        }

return result;                                          /* return the result of the attach */
}


/* Detach a disc image file from a unit.

   The file currently attached to the unit specified by the "uptr" parameter is
   detached.  This is the simulation equivalent of unloading the disc pack from
   the drive.  An unattached unit reports Not Ready error status.
*/

static t_stat da_detach (UNIT *uptr)
{
t_stat result;
const int32 unit = uptr - da_unit;                      /* the number of the detaching unit */

result = dl_detach (&icd_cntlr [unit], uptr);           /* detach the drive */

if (result == SCPE_OK)                                  /* if the detach was successful */
    set_response (unit, CLEAR);                         /*   then clear its PPR, as it's no longer ready */

return result;                                          /* return the result of the detach */
}


/* 7906H/20H/25H disc bootstrap loader (12992H).

   The HP 1000 uses the 12992H boot loader ROM to bootstrap the ICD discs.  Bit
   12 of the S register determines whether an RPL or manual boot is performed.
   Bits 1-0 specify the head number to use.

   The loader reads 256 words from cylinder 0 sector 0 of the specified head
   into memory starting at location 2011 octal.  Loader execution ends with one
   of the following instructions:

     * HLT 11     - the drive aborted the transfer due to an unrecoverable error
     * JSB 2055,I - the disc read succeeded

   The ICD drives are not supported on the 2100/14/15/16 CPUs, so no 21xx loader
   is provided.
*/

static const LOADER_ARRAY da_loaders = {
    {                               /* HP 21xx Loader does not exist */
      IBL_NA,                       /*   loader starting index */
      IBL_NA,                       /*   DMA index */
      IBL_NA,                       /*   FWA index */
      { 0 } },

    {                               /* HP 1000 Loader ROM (12992H) */
      IBL_START,                    /*   loader starting index */
      IBL_DMA,                      /*   DMA index */
      IBL_FWA,                      /*   FWA index */
      { 0102501,                    /*   77700:  START LIA 1         GET SWITCH REGISTER SETTING */
        0100044,                    /*   77701:        LSL 4         SHIFT A LEFT 4 */
        0006111,                    /*   77702:        CLE,SLB,RSS   SR BIT 12 SET FOR MANUAL BOOT? */
        0100041,                    /*   77703:        LSL 1         NO, SHIFT HEAD # FOR RPL BOOT */
        0001424,                    /*   77704:        ALR,ALR       SHIFT HEAD 2, CLEAR SIGN */
        0033744,                    /*   77705:        IOR HDSEC     SET EOI BIT */
        0073744,                    /*   77706:        STA HDSEC     PLACE IN COMMAND BUFFER */
        0017756,                    /*   77707:        JSB BTCTL     SEND DUMMY,U-CLR,PP */
        0102510,                    /*   77710:        LIA IBI       READ INPUT REGISTER */
        0101027,                    /*   77711:        ASR 7         SHIFT DRIVE 0 RESPONSE TO LSB */
        0002011,                    /*   77712:        SLA,RSS       DID DRIVE 0 RESPOND? */
        0027710,                    /*   77713:        JMP *-3       NO, GO LOOK AGAIN */
        0107700,                    /*   77714:        CLC 0,C       */
        0017756,                    /*   77715:        JSB BTCTL     SEND TALK, CL-RD,BUS HOLDER */
        0002300,                    /*   77716:        CCE           */
        0017756,                    /*   77717:        JSB BTCTL     TELL CARD TO LISTEN */
        0063776,                    /*   77720:        LDA DMACW     LOAD DMA CONTROL WORD */
        0102606,                    /*   77721:        OTA 6         OUTPUT TO DCPC */
        0106702,                    /*   77722:        CLC 2         READY DCPC */
        0063735,                    /*   77723:        LDA ADDR1     LOAD DMA BUFFER ADDRESS */
        0102602,                    /*   77724:        OTA 2         OUTPUT TO DCPC */
        0063740,                    /*   77725:        LDA DMAWC     LOAD DMA WORD COUNT */
        0102702,                    /*   77726:        STC 2         READY DCPC */
        0102602,                    /*   77727:        OTA 2         OUTPUT TO DCPC */
        0103706,                    /*   77730:        STC 6,C       START DCPC */
        0102206,                    /*   77731:  TEST  SFC 6         SKIP IF DMA NOT DONE */
        0117750,                    /*   77732:        JSB ADDR2,I   SUCCESSFUL END OF TRANSFER */
        0102310,                    /*   77733:        SFS IBI       SKIP IF DISC ABORTED TRANSFER */
        0027731,                    /*   77734:        JMP TEST      RECHECK FOR TRANSFER END */
        0102011,                    /*   77735:  ADDR1 HLT 11B       ERROR HALT */
        0000677,                    /*   77736:  UNCLR OCT 677       UNLISTEN */
        0000737,                    /*   77737:        OCT 737       UNTALK */
        0176624,                    /*   77740:  DMAWC OCT 176624    UNIVERSAL CLEAR,LBO */
        0000440,                    /*   77741:  LIST  OCT 440       LISTEN BUS ADDRESS 0 */
        0000550,                    /*   77742:  CMSEC OCT 550       SECONDARY GET COMMAND */
        0000000,                    /*   77743:  BOOT  OCT 0         COLD LOAD READ COMMAND */
        0001000,                    /*   77744:  HDSEC OCT 1000      HEAD,SECTOR PLUS EOI */
        0000677,                    /*   77745:  UNLST OCT 677       ATN,PRIMARY UNLISTEN,PARITY */
        0000500,                    /*   77746:  TALK  OCT 500       SEND READ DATA */
        0100740,                    /*   77747:  RDSEC OCT 100740    SECONDARY READ DATA */
        0102055,                    /*   77750:  ADDR2 OCT 102055    BOOT EXTENSION STARTING ADDRESS */
        0004003,                    /*   77751:  CTLP  OCT 4003      INT=LBO,T,CIC */
        0000047,                    /*   77752:        OCT 47        PPE,L,T,CIC */
        0004003,                    /*   77753:        OCT 4003      INT=LBO,T,CIC */
        0000413,                    /*   77754:        OCT 413       ATN,P,L,CIC */
        0001015,                    /*   77755:        OCT 1015      INT=EOI,P,L,CIC */
        0000000,                    /*   77756:  BTCTL NOP           */
        0107710,                    /*   77757:        CLC IBI,C     RESET IBI */
        0063751,                    /*   77760:  BM    LDA CTLP      LOAD CONTROL WORD */
        0102610,                    /*   77761:        OTA IBI       OUTPUT TO CONTROL REGISTER */
        0102710,                    /*   77762:        STC IBI       RETURN IBI TO DATA MODE */
        0037760,                    /*   77763:        ISZ BM        INCREMENT CONTROL WORD POINTER */
        0002240,                    /*   77764:        SEZ,CME       */
        0127756,                    /*   77765:        JMP BTCTL,I   RETURN */
        0063736,                    /*   77766:  LABL  LDA UNCLR     LOAD DATA WORD */
        0037766,                    /*   77767:        ISZ LABL      INCREMENT WORD POINTER */
        0102610,                    /*   77770:        OTA IBI       OUTPUT TO HPIB */
        0002021,                    /*   77771:        SSA,RSS       SKIP IF LAST WORD */
        0027766,                    /*   77772:        JMP LABL      GO BACK FOR NEXT WORD */
        0102310,                    /*   77773:        SFS IBI       SKIP IF LAST WORD SENT TO BUS */
        0027773,                    /*   77774:        JMP *-1       RECHECK ACCEPTANCE */
        0027757,                    /*   77775:        JMP BTCTL+1   */
        0000010,                    /*   77776:  DMACW ABS IBI       */
        0170100 } }                 /*   77777:        ABS -START    */
    };

/* Device boot routine.

   This routine is called directly by the BOOT DA{n} and LOAD DA{n} commands to
   copy the device bootstrap into the upper 64 words of the logical address
   space.  It is also called indirectly by a BOOT CPU or LOAD CPU command when
   the specified HP 1000 loader ROM socket contains a 12992H ROM.

   When called in response to a BOOT DA{n} or LOAD DA{n} command, the "unitno"
   parameter indicates the unit number specified or is -1 if the device name is
   specified, and "dptr" points at the DA device structure.  The bootstrap
   supports loading only from the disc at bus address 0 only.  The 12992F loader
   ROM will be copied into memory and configured for the DA select code.  The S
   register will be set as it would be by the front-panel microcode.

   When called for a BOOT/LOAD CPU command, the "unitno" parameter indicates the
   select code to be used for configuration, and "dptr" will be NULL.  As above,
   the 12992H loader ROM will be copied into memory and configured for the
   specified select code.  The S register is assumed to be set correctly on
   entry and is not modified.

   The loader expects the S register to be is set as follows:

      15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | ROM # | 0   1 |      select code      |   reserved    | head  |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Bit 12 must be 1 for a manual boot.  Bits 5-2 are nominally zero but are
   reserved for the target operating system.  For example, RTE uses bit 5 to
   indicate whether a standard (0) or reconfiguration (1) boot is desired.

   The boot routine sets bits 15-6 of the S register to appropriate values.
   Bits 5-3 and 1-0 retain their original values, so S should be set before
   booting.  These bits are typically set to 0, although bit 5 is set for an RTE
   reconfiguration boot, and bits 1-0 may be set if booting from a head other
   than 0 is desired.
*/

static t_stat da_boot (int32 unitno, DEVICE *dptr)
{
static const HP_WORD da_preserved   = 0000073u; /* S-register bits 5-3 and 1-0 are preserved */
static const HP_WORD da_manual_boot = 0010000u; /* S-register bit 12 set for a manual boot */
uint32 status;

if (dptr == NULL)                                           /* if we are being called for a BOOT/LOAD CPU */
    status = cpu_copy_loader (da_loaders, unitno,           /*   then copy the boot loader to memory */
                              IBL_S_NOCLEAR, IBL_S_NOSET);  /*     but do not alter the S register */

else if (unitno == -1                                           /* otherwise if the device was specified */
  || ADDRESS (da_unit [unitno].flags) == 0)                     /*   or the unit is at bus address 0 */
    status = cpu_copy_loader (da_loaders, da_dib.select_code,   /*     then copy the boot loader to memory */
                              da_preserved, da_manual_boot);    /*       and configure the S register if 1000 CPU */

else                                                    /* otherwise */
    status = 0;                                         /*   reject other addresses as unsupported */

if (status == 0)                                        /* if the copy failed */
    return SCPE_NOFNC;                                  /*   then reject the command */
else                                                    /* otherwise */
    return SCPE_OK;                                     /*   the boot loader was successfully copied */
}


/* Load or unload a unit's heads.

   The heads are automatically loaded when a unit is attached and unloaded when
   a unit is detached.  While a unit is attached, the heads may be manually
   unloaded; this yields a "not ready" status if the unit is accessed.  An
   unloaded drive may be manually loaded, returning the unit to "ready" status.

   The ICD controller sets Drive Attention status when the heads unload and also
   asserts a parallel poll response if the heads unload while in idle state 2
   (i.e., after an End command).


   Implementation notes:

    1. The 13365 manual says on page 28 that Drive Attention status is
       "Generated whenever...the drive unloads and the controller is in Idle
       State 2 or 3."  However, the ICD diagnostic tests for Drive Attention
       status on head unload immediately after the Request Status command that
       completes the previous step, which leaves the controller in idle state 1.

       Moreover, the diagnostic does NOT check for Drive Attention status if the
       Amigo ID is 2 (MAC controller).  But the 12745 manual says on page 3-7
       that the status is returned if "...Drive becomes not ready (heads
       unload)" with no mention of controller state.

       It appears as though the diagnostic test is exactly backward.  However,
       we match the diagnostic expectation below.
*/

static t_stat da_load_unload (UNIT *uptr, int32 value, char *cptr, void *desc)
{
const int32 unit = uptr - da_unit;                      /* the number of the referenced unit */
const t_bool load = (value != UNIT_UNLOAD);             /* true if the heads are loading */
t_stat result;

result = dl_load_unload (&icd_cntlr [unit], uptr, load);    /* load or unload the heads */

if (result == SCPE_OK && not load) {                    /* was the unload successful? */
    icd_cntlr [unit].status = drive_attention;          /* set Drive Attention status */

    if (uptr->Op == End)                                /* is the controller in idle state 2? */
        set_response (unit, SET);                       /* enable PPR */
    }

return result;                                          /* return the result of the load or unload */
}


/* Set a device bus address.

   This validation routine is called to set the bus address of a simulated unit.
   Bus addresses range from 0-7.  Each disc drive attached to the bus is
   assigned a unique HP-IB bus address.  Addressing conflicts are permitted
   during assignment but are checked before execution begins, and invalid
   combinations must be corrected.

   On entry, "uptr" points at the simulator unit to configure, "cptr" points to
   the number to parse, and "value" and "desc" are not used.


   Implementation notes:

    1. The alternate of validating all changes at assignment time and rejecting
       any assignment that conflicts would force the user to make intermediate
       assignments to swap a pair of bus addresses.  It would also require
       trapping unit enable commands to prevent enabling a unit whose
       assignment conflicts with an existing one.  These same reasons are why
       setting conflicting select code assignments are allowed, and a
       consistency check is made prior to commencing execution.
*/

static t_stat da_set_address (UNIT *uptr, int32 value, char *cptr, void *desc)
{
uint32 new_value;
t_stat status;

if (cptr == NULL)                                           /* if the value is not present */
    return SCPE_MISVAL;                                     /*   then report a missing value */
else                                                        /* otherwise */
    new_value = (uint32) get_uint (cptr, 10, 7, &status);   /*   parse the new value */

if (status == SCPE_OK)                                  /* if the parse succeeded */
    uptr->flags = uptr->flags & ~UNIT_ADDRESS           /*   then change the bus address in the unit flags */
                    | TO_ADDRESS (new_value);

return status;                                          /* return the result of the parse */
}


/* Show a device bus address.

   This display routine is called to show the bus address for a simulator unit.
   The output stream is passed in the "st" parameter, a pointer to the simulator
   unit is passed in "uptr", and the other parameters are not used.
*/

static t_stat da_show_address (FILE *st, UNIT *uptr, int32 value, void *desc)
{
fprintf (st, "bus=%d", ADDRESS (uptr->flags));          /* display the bus address */

return SCPE_OK;                                         /* address reports always succeed */
}



/* Amigo global bus routines.

   The following bus routines are called by the disc interface to initialize the
   bus, supply a data byte to the bus, and allow the devices to respond to
   control line changes on the bus.

   This protocol module maintains the current state of the channel, so that the
   routines know which devices are present and addressed.  This is more
   efficient than querying each bus address at each call to determine whether a
   device is present and, if so, whether it is presently addressed to talk or
   listen.  The channel state consists of bus address bitmaps of the devices
   present on the bus, the devices currently listening, and the devices' current
   parallel poll responses, the address of the current talker, if any, and flags
   to indicate whether a bus transaction is in progress (needed to decide if an
   Untalk or Unlisten is aborting a transaction) and whether a parallel poll is
   currently in progress (needed in case a device's poll response changes during
   a poll).

   In simulation, changes in control line state are communicated to all modules
   via control/respond function calls, and data is exchanged between talker and
   listeners via source/acceptor function calls.  Each protocol module receives
   one acceptor or responder call per bus state change and is responsible for
   distributing these notifications to all devices under its control.  This
   arrangement accommodates multiple protocol handlers per bus, although this
   feature is not used on RTE systems.
*/


/* Initialize the bus.

   This bus routine is called during I/O initialization each time machine
   instruction execution is to begin.  The routine returns TRUE to permit
   execution to commence or FALSE to inhibit execution.  Bus initialization is
   optional for modules but is required in our case to check the SIMH unit
   assignments for bus address conflicts.

   On entry, the "card" parameter is the identifier of the disc interface to
   which devices using this module's protocol are connected.  It is not used in
   our case because RTE permits only one protocol per card.

   Conflicts are detected by setting up a table of bus addresses and counting
   the number of SIMH units for that apply to each.  This table is then used to
   verify that each SIMH unit has a unique bus address.  Each table entry is a
   conflict counter that records the number of SIMH units assigned to the
   specific bus address.

   First, the conflict table is initialized to all zeros, and the bitmap of
   present devices is cleared.  Then each enabled SIMH unit is examined
   (disabled units are "not present"), and the assigned bus address is
   determined.  Then the conflict table entry for the specified address is
   incremented, and the bus address of the device is entered into the device
   presence bitmap.

   Once all of the SIMH units have been entered into the conflict table, the
   table is then scanned in bus address order for problems.  If a unit's
   reference counter is greater than one, "bus address has multiple assignments"
   is reported.

   Finally, if any conflicts exist, the routine returns FALSE to inhibit machine
   instruction execution.


   Implementation notes:

    1. In addition to being called as a result of a BOOT, RUN, GO, or CONTINUE
       command, we are also called after any concurrent-mode command, as that
       suspends and then resumes execution.

    2. Channel presence is recalculated to remove devices disabled by the user.
*/

t_bool da_bus_initialize (CARD_ID card)
{
UNIT   *uptr;
uint32 address;
uint8  conflicts [DEVICE_COUNT];
t_bool coherent = TRUE;

memset (conflicts, 0, sizeof conflicts);                /* initialize the conflict table */

channel.present = 0;                                    /* reset the map of installed devices */

for (uptr = da_unit; uptr < &da_unit [DA_UNITS]; uptr++)    /* loop through the SIMH units */
    if (not (uptr->flags & UNIT_DIS)) {                     /* if the unit is not disabled */
        address = ADDRESS (uptr->flags);                    /*   then get the address */

        conflicts [address]++;                          /* bump the assignment count for this unit */
        channel.present |= 1u << address;               /*   and set the "device is present" bit */
        }

for (address = 0; address < DEVICE_COUNT; address++)    /* loop through each device */
    if (conflicts [address] > 1) {                      /* if the address has multiple SIMH units */
        coherent = FALSE;                               /*   then the assignments are not coherent */

        cprintf ("DA bus address %u has multiple assignments\n",
                 address);
        }

return coherent;                                        /* return the bus coherency state */
}


/* Accept a command or data byte from the bus.

   This routine is called whenever a controller or talker sources a data byte
   onto the bus.  It is responsible for decoding primary, secondary, talk, and
   listen commands, and for calling routines to process secondaries and incoming
   data.  It returns TRUE if any device on the bus accepted the byte and FALSE
   if there are no acceptors.

   On entry, the "card" parameter is the identifier of the disc interface to
   which devices using this module's protocol are connected, and the "bus"
   parameter is the set of eight control and eight data lines representing the
   HP-IB.  The card ID is not used in our case because RTE permits only one
   protocol per card.

   Devices connected to the bus specified by the "card" parameter are offered
   a data byte and associated signals specified by the "bus" parameter.  If
   there are no devices present on the bus, then there are no acceptors, and the
   routine returns FALSE.  Otherwise, the state of the ATN bus signal is
   examined to determine if the byte is a command byte (ATN asserted) or data
   byte (ATN denied).

   Primary commands placed on the bus may be universal (applying to all present
   devices) or addressed (applying only to those devices that are addressed to
   listen).  Acceptance is contingent on whether the command is universal
   (always accepted) or addressed (accepted only if someone is listening).

   Address commands select the device that will talk and the device(s) that will
   listen.  If the addressed device is present, the byte is accepted; acceptance
   also occurs if the UNT (untalk) UNL (unlisten) commands are received.  A talk
   or listen command directed to a device not present is not accepted.

   Secondary commands and data bytes are accepted only if one or more listeners
   are present.

   This routine is called once for each byte sourced and is responsible for
   notifying all connected devices on the specified bus.  For universal and
   addressed commands and data, the routine iterates through the list of present
   or currently listening devices, passing the bus state to each in turn.  The
   type of byte passed is determined by the state of the ATN signal and, if ATN
   is asserted, by the high-order bits of the value.

   If a Talk, Unlisten, or Untalk command is received while a command is
   currently in progress, a channel abort occurs.  This terminates the current
   command with an I/O Program Error.  A new Listen command does not cause a
   channel abort, because the current listener remains addressed.

   For secondaries, the Amigo Identify command sequence is decoded here, as it
   does not follow the standard HP-IB protocol.  Otherwise, secondaries are
   processed by each device that is addressed to listen or talk.

   If ATN is not asserted, a data byte has been sourced.  It is processed by
   each device that is addressed to listen.

   The Amigo command sequence does not provide a byte count for disc read and
   write commands, so the controller continues to source or accept data bytes
   until the device is unaddressed.  Normally, this is done by an Unlisten or
   Untalk.  However, per IEEE-488, a listening device may be unaddressed by IFC,
   by an Unlisten, or by addressing the device to talk, and a talking device may
   be unaddressed by IFC, by addressing another device to talk (or no device via
   Untalk), or by addressing the device to listen.  Therefore, we must keep
   track of whether the unit stopped talking or listening, and if it has, we
   check for command termination.

   If the controller is unaddressed in the middle of a sector transfer, the read
   or write must be terminated cleanly to ensure that the disc image is
   coherent.  It is also permissible to untalk the controller before all of the
   requested status bytes are returned.

   In addition, the controller has no way to inform the host that an error has
   occurred that prevents the command from continuing.  For example, if a data
   error is encountered while reading or a protected track is encountered while
   writing, the controller must still source or sink data bytes until the
   command is terminated by the host.  The controller handles read errors by
   sourcing a single data byte tagged with EOI and write errors by sinking data
   bytes until EOI is seen or the unit is unaddressed.

   Therefore, if the unit is unaddressed while a read, write, or status command
   is transferring data, the unit service must be scheduled to end the current
   command.  Unaddressing while an error condition is present merely terminates
   the source or sink operation.


   Implementation notes:

    1. The 13365 manual does not indicate that the controller responds to
       Universal Clear, but the 12992H loader ROM issues this primary and
       expects the controller to initialize.

    2. The Identify command is an Untalk immediately followed by a secondary
       address.  This sequence requires knowing that an Untalk was the last
       primary command received, so we save it before returning (it must be
       saved globally to ensure that it persists across a SAVE/RESTORE command
       sequence).  If a secondary is immediately preceded by an Untalk, and the
       address contained in the secondary is present, then an Identify command
       has been received.  We must then mark the device as the current talker so
       that the bus responder can reschedule the proper SIMH unit when the
       controller denies ATN to permit return of the identity bytes.

    3. It is not necessary to check for listening when processing addressed
       commands, as only listeners are called by the bus source.
*/

t_bool da_bus_accept (CARD_ID card, BUS_CABLE bus)
{
const uint32 address = BUS_ADDRESS (bus);
const uint32 address_bit = 1u << address;
uint32       bus_address, dev_map;
uint32       unlisteners = 0;
uint32       untalker = BUS_UNADDRESS;
t_bool       accepted = TRUE;
t_bool       addressed = FALSE;

if (not channel.present)                                /* if there are no devices present */
    return FALSE;                                       /*   then no one accepted */

else if (bus & Bus_ATN) {                               /* otherwise if this is a command */
    switch (BUS_GROUP (bus)) {                          /*   then dispatch the bus group */

        case BUS_PCG:                                   /* primary command group */
            dev_map = channel.listeners;                /* get the current set of devices listening */

            switch (address) {

                case 0x14:                              /* Universal Device Clear */
                    dev_map = channel.present;          /* clear all devices present on the bus */

                /* fall through into the SDC case */

                case 0x04:                              /* Selected Device Clear */
                case 0x05:                              /* SDC with parity freeze */
                    tprintf_bus (da_dev, bus, "%s device clear\n",
                                 (address == 0x014 ? "Universal" : "Selected"));

                    for (bus_address = 0; dev_map != 0; bus_address++) {
                        if (dev_map & 1) {                                  /* if the device is present */
                            sim_cancel (&da_unit [bus_address]);            /* cancel any in-progress command */
                            dl_idle_controller (&icd_cntlr [bus_address]);  /* idle the controller */
                            if_dsj [bus_address] = 0;                       /* clear DSJ */
                            if_state [bus_address] = idle;                  /* idle the interface */
                            set_response (bus_address, SET);                /* enable PPR */

                            tprintf (da_dev, TRACE_CMD, "Unit %d cleared\n", bus_address);
                            }

                        dev_map >>= 1;
                        }
                    break;


                default:                                /* unsupported universal command */
                    break;                              /* universals are always accepted */
                }

            break;


        case BUS_LAG:                                   /* listen address group */
            if (address == BUS_UNADDRESS) {             /* if this is an Unlisten */
                unlisteners = channel.listeners;        /*   then save the list of previous listeners */
                channel.listeners = 0;                  /*     and then clear all of the listeners */

                tprintf_bus (da_dev, bus, "Unlisten\n");
                }

            else if (channel.present & address_bit) {   /* otherwise if the addressed device is present */
                channel.listeners |= address_bit;       /*   then it becomes a listener */

                if (channel.talker == address) {        /* if it is the current talker */
                    untalker = channel.talker;          /*   then save the previous talker */
                    channel.talker = BUS_UNADDRESS;     /*     and then untalk it */
                    }

                addressed = TRUE;                       /* unit is now addressed */

                tprintf_bus (da_dev, bus, "Listen %u\n", address);
                }

            else                                        /* other listen addresses */
                accepted = FALSE;                       /*   are not accepted */

            break;


        case BUS_TAG:                                   /* talk address group */
            if (address == BUS_UNADDRESS) {             /* if this is an Untalk */
                untalker = channel.talker;              /*   then save the previous talker */
                channel.talker = address;               /*     and then clear the talker */

                tprintf_bus (da_dev, bus, "Untalk\n");
                }

            else if (channel.present & address_bit) {   /* otherwise if the addressed device is present */
                channel.talker = address;               /*   then it becomes the current talker */

                if (channel.listeners & address_bit) {  /* if the new talker had been listening */
                    unlisteners = address_bit;          /*   than save the previous listener */
                    channel.listeners &= ~address_bit;  /*     and clear its listener bit */
                    }

                addressed = TRUE;                       /* the unit is now addressed */

                tprintf_bus (da_dev, bus, "Talk %u\n", address);
                }

            else                                        /* other talk addresses */
                accepted = FALSE;                       /*   are not accepted */

            break;


        case BUS_SCG:                                   /* Secondary Command Group */
            if (channel.command == BUS_UNTALK           /* if preceded by an Untalk */
              && channel.present & address_bit) {       /*   and the addressed device is present */
                if_command [address] = amigo_identify;  /*     then this is an Amigo ID sequence */
                if_state [address] = command_exec;      /*       so set it up for execution */

                channel.talker = address;               /* temporarily address the device to talk */
                icd_cntlr [address].index = 0;          /*   and reset the buffer index */

                tprintf_bus (da_dev, bus, "Identify %u\n", address);

                da_unit [address].wait = icd_cntlr [address].cmd_time;  /* schedule the unit */
                activate_unit (&da_unit [address]);                     /*   to execute the command */

                tprintf (da_dev, TRACE_INCO, "Unit %d %s command initiated\n",
                         address, if_command_name [amigo_identify]);
                }

            else {                                      /* otherwise */
                dev_map = channel.listeners;            /*   get the current set of listeners */

                if (channel.talker != BUS_UNADDRESS)    /* if a talker is addressed */
                    dev_map |= 1u << channel.talker;    /*   then include it in the set */

                accepted = (dev_map != 0);              /* TRUE if some device receives the secondary */

                for (bus_address = 0; dev_map != 0; bus_address++) {
                    if (dev_map & 1)                            /* if the device is addressed */
                        process_secondary (bus_address, bus);   /*   then process the secondary */

                    dev_map >>= 1;                      /* move to the next device */
                    }
                }
            break;                                      /* end of secondary processing */
        }

    channel.command = BUS_COMMAND (bus);                /* save the command for a potential Identify */

    if (addressed && sim_is_active (&da_unit [address])) {  /* is the unit being addressed while it is busy? */
        if_state [address] = command_wait;                  /* change the interface state to wait */
        di_bus_control (da, address, Bus_NRFD, 0);          /*   and assert NRFD to hold off the card */

        tprintf (da_dev, TRACE_INCO, "Unit %d addressed while controller is busy\n", address);
        }

    if (unlisteners != 0)                                           /* if some unit was unlistened */
        for (bus_address = 0; unlisteners != 0; bus_address++) {
            if (unlisteners & 1) {                                  /* if the device was addressed */
                if (icd_cntlr [bus_address].state == cntlr_busy)    /* if the controller is busy, */
                    complete_write (bus_address);                   /*   then check for write completion */

                else if (if_command [bus_address] == invalid)   /* if a command was aborting, */
                    complete_abort (bus_address);               /*   then complete it */

                else if (if_state [bus_address] == opcode_wait              /* if waiting for an opcode */
                  || if_state [bus_address] == parameter_wait)              /*   or a parameter, */
                    abort_command (bus_address, io_program_error, idle);    /*     then abort the pending command */

                if (da_unit [bus_address].wait > 0)         /* was service requested? */
                    activate_unit (&da_unit [bus_address]); /* schedule the unit */
                }

            unlisteners >>= 1;
            }

    else if (untalker != BUS_UNADDRESS) {               /* if the unit was untalked */
        if (icd_cntlr [untalker].state == cntlr_busy)   /*   then if the controller is busy, */
            complete_read (untalker);                   /*     then check for read completion */

        else if (if_command [untalker] == invalid)      /* if a command was aborting, */
            complete_abort (untalker);                  /*   then complete it */

        if (da_unit [untalker].wait > 0)                /* was service requested? */
            activate_unit (&da_unit [untalker]);        /* schedule the unit */
        }
    }                                                   /* end of bus command processing */

else {                                                  /* otherwise it is bus data (ATN is denied) */
    dev_map = channel.listeners;                        /* get the current set of devices listening */

    accepted = (dev_map != 0);                          /* TRUE if some device receives the data byte */

    for (bus_address = 0; dev_map != 0; bus_address++) {
        if (dev_map & 1)                                /* if the device is present */
            process_data (bus_address, bus);            /*   then process the data byte */

        dev_map >>= 1;                                  /* move to the next device */
        }
    }

return accepted;                                        /* indicate the acceptance condition */
}


/* Respond to a bus control line change.

   This routine is called whenever the state of the bus control lines change
   outside of a data source/accept handshake.  Any devices affected are notified
   of the change.  The routine returns the current set of parallel poll
   responses.

   On entry, the "card" parameter is the identifier of the disc interface to
   which devices using this module's protocol are connected, and the "new_cntl"
   parameter is the set of eight control lines representing the new state of the
   HP-IB.  The card ID is not used in our case because RTE permits only one
   protocol per card.

   There are three conditions to which we must respond:

    1. A parallel poll is being conducted.  We record the poll state, so that if
       a poll response is subsequently enabled while a poll is still in
       progress, we can notify the controller.

    2. An Interface Clear is initiated.  IFC unaddresses all units, so any
       transactions in progress must be terminated as though an Untalk and
       Unlisten were accepted from the bus.

    3. Attention or Not Ready for Data are denied.  A device addressed to talk
       must wait for ATN to deny before data may be sent.  Also, a listener that
       has asserted NRFD must deny it before a talker may send data.  If a
       talker is sending data and both ATN and NRFD are denied, then we
       reenter the service routine to send the next byte.


   Implementation notes:

    1. Amigo Identify is handled as a special case because of the odd bus
       protocol employed (the device puts data on the bus without having been
       addressed as the talker).  When the command is processed, the device is
       temporarily assigned as the talker for the duration of the command, which
       is preceded by an Untalk.  This allows us to know which unit to
       reschedule if ATN or NRFD is asserted when the Identify is serviced.

    2. We return the parallel poll response set regardless of whether or not a
       poll is in progress.  If the controller is not polling, the returned
       response will be ignored.
*/

BUS_DATA da_bus_respond (CARD_ID card, BUS_CNTL new_cntl)
{
uint32 dev_map, address;

channel.poll_active = ((new_cntl & BUS_PPOLL) == BUS_PPOLL);    /* record the parallel poll state */

if (new_cntl & Bus_IFC) {                               /* if interface clear is asserted */
    channel.listeners = 0;                              /*   then unaddress all listeners */
    channel.talker = BUS_UNADDRESS;                     /*     and unaddress the current talker */

    dev_map = channel.present;                          /* clear all devices present on the bus */

    for (address = 0; dev_map != 0; address++) {            /* for each device */
        if (dev_map & 1)                                    /* if the device is present */
            if (icd_cntlr [address].state == cntlr_busy) {  /* is the controller busy? */
                complete_write (address);                   /* check for write completion */
                complete_read (address);                    /*   or read completion */

                if (da_unit [address].wait > 0)         /* is service needed? */
                    activate_unit (&da_unit [address]); /* activate the unit */
                }

            else if (if_command [address] == invalid)   /* if a command was aborting, */
                complete_abort (address);               /*   then complete it */

            else if (if_state [address] == opcode_wait              /* if we're waiting for an opcode */
              || if_state [address] == parameter_wait)              /*   or a parameter, */
                abort_command (address, io_program_error, idle);    /*   then abort the pending command */

        dev_map >>= 1;
        }
    }

if (di_bus_ready (card)                                 /* if the card is ready for data */
  && channel.talker != BUS_UNADDRESS                    /*   and a talker is addressed */
  && (if_state [channel.talker] == read_xfer            /*     and the interface waiting to send data */
  || if_state [channel.talker] == error_source))        /*       or source error bytes */
    da_service (&da_unit [channel.talker]);             /*         then start or resume the transfer */

return channel.poll_response;                           /* return the current poll response set */
}



/* Amigo disc local utility routines */


/* Process a secondary command.

   This routine is called to process a secondary command received from the bus
   for an addressed device.  Each device present on the bus that has been
   addressed either as a listener or talker receives the secondary.  Secondaries
   are used by the Amigo protocol to identify various types of messages,
   specifically:

     T/L  Value  Identifies
     ---  -----  ----------------------------
      L    00H   Receive Write Data
      L    08H   Receive Disc Command
      L    09H   Cyclic Redundancy Check
      L    10H   Secondary Hard Clear
      L    1FH   Initiate Self-Test
      L    1EH   Write Loopback Record

      T    00H   Send Read Data
      T    08H   Send Disc Status
      T    09H   Cyclic Redundancy Check
      T    10H   Return Device-Specified Jump
      T    1EH   Read Loopback Record
      T    1FH   Return Self-Test Result

   On entry, the "address" parameter contains the bus address of the target
   device, which will be either a listener or talker.  The "bus" parameter
   contains the set of eight control and eight data lines representing the HP-IB
   state.

   Most of the work involves decoding secondary commands and their associated
   data parameters.  The interface state is changed as needed to track the
   command protocol.  The states processed in this routine are:

     opcode_wait
     -----------
     A Receive Disc Command secondary has been received, and the interface is
     waiting for the opcode that should follow.

     parameter_wait
     --------------
     A disc opcode or interface command has been received, and the interface is
     waiting for a parameter byte that should follow.

     write_wait
     ----------
     A disc write command has been received, and the interface is waiting for
     the Receive Write Data secondary that should follow.

     read_wait
     ---------
     A disc read command has been received, and the interface is waiting for the
     Send Read Data secondary that should follow.

     status_wait
     -----------
     A disc status command has been received, and the interface is waiting for
     the Send Disc Status secondary that should follow.

     write_xfer
     ----------
     A disc write is in progress, and the interface is waiting for a data byte
     that should follow.

     error_sink
     ----------
     A disc write has terminated with an error, and the interface is waiting to
     absorb all of the remaining data bytes of the transfer.

   Disc commands and parameters are assembled in the sector buffer before being
   passed to the disc library to start the command.  Once the command is
   started, the interface state is set either to execute the command or to wait
   for the receipt of a data transfer secondary before executing, depending on
   the command.

   Two disc command protocol errors are detected.  First, an Illegal Opcode is
   identified during the check for the expected number of disc command
   parameters.  This allows us to sink an arbitrary number of parameter bytes.
   Second, an I/O Program Error occurs if an unsupported secondary is received
   or the HP-IB sequence is incorrect.  The latter occurs if a command has the
   wrong number of parameters or a secondary data transfer sequence is invalid.

   Disc commands that require data transfers (e.g., Read, Write, Request Status)
   involve a pair of secondaries.  The first transmits the command, and the
   second transmits or receives the data.  If one occurs without the other, an
   I/O Program Error occurs.

   A secondary or command that generates an I/O Program Error is always ignored.
   Error recovery is as follows:

    - An unsupported talk secondary sends a single data byte tagged with EOI.

    - An unsupported listen secondary accepts and discards any accompanying data
      bytes until EOI is asserted or an Unlisten is received.

    - A supported command with too few parameter bytes or for which the last
      parameter byte is not tagged with EOI (before unlisten) does nothing.

    - A supported command with too many parameter bytes accepts and discards
      excess parameter bytes until EOI is asserted or an Unlisten is received.

    - A read or status command that is not followed by a Send Read Data or a
      Send Disc Status secondary does nothing.  The unexpected secondary is
      executed normally.

    - A write command that is not followed by a Receive Write Data secondary
      does nothing.  The unexpected secondary is executed normally.

    - A Send Read Data or a Send Disc Status secondary that is not preceded by a
      read or status command sends a single data byte tagged with EOI.

    - A Receive Write Data secondary that is not preceded by a write command
      accepts and discards data bytes until EOI is asserted or an Unlisten is
      received.
*/

static void process_secondary (uint32 address, BUS_CABLE bus)
{
const uint32   secondary  = BUS_SECONDARY (bus);        /* the secondary value without parity */
const IF_STATE last_state = if_state [address];         /* the interface state at entry */
t_bool         initiated  = FALSE;
t_bool         accepted   = TRUE;

icd_cntlr [address].index = 0;                          /* reset the buffer index */

if (channel.listeners & 1u << address) {                /* if this device is listening */
    if (if_state [address] == write_wait                /*   then if we're waiting for a write data secondary */
      && secondary != 0x00)                             /*     but it's not there */
        abort_command (address, io_program_error,       /*       then abort the pending command */
                       idle);                           /*         and process the new command */

    switch (secondary) {                                /* dispatch the listen secondary */

        case 0x00:                                      /* Receive Write Data */
            tprintf_bus (da_dev, bus, "Receive Write Data\n");

            if (if_state [address] != write_wait)           /* if we're not expecting it */
                abort_command (address, io_program_error,   /*   abort and sink any data */
                               error_sink);
            else {                                                      /* the sequence is correct */
                if_state [address] = command_exec;                      /* the command is ready to execute */
                da_unit [address].wait = icd_cntlr [address].cmd_time;  /* schedule the unit */
                di_bus_control (da, address, Bus_NRFD, 0);              /* assert NRFD to hold off the card */
                }

            initiated = TRUE;                           /* log the command or abort initiation */
            break;


        case 0x08:                                      /* disc commands */
            tprintf_bus (da_dev, bus, "Receive Disc Command\n");

            if_command [address] = disc_command;        /* set the command and wait */
            if_state [address] = opcode_wait;           /*  for the opcode that must follow */
            break;


        case 0x09:                                      /* CRC (Listen) */
            tprintf_bus (da_dev, bus, "Receive CRC\n");

            if_command [address] = crc_listen;          /* set up the command */
            if_state [address] = error_sink;            /* sink any data that will be coming */
            initiated = TRUE;                           /* log the command initiation */
            break;


        case 0x10:                                      /* Amigo Clear */
            tprintf_bus (da_dev, bus, "Amigo Clear\n");

            if_command [address] = amigo_clear;         /* set up the command */
            if_state [address] = parameter_wait;        /* a parameter must follow */
            icd_cntlr [address].length = 1;             /* set to expect one (unused) byte */
            break;


        case 0x1E:                                      /* Write Loopback */
            tprintf_bus (da_dev, bus, "Write Loopback Record\n");

            if_command [address] = write_loopback;      /* set up the command */
            if_state [address] = write_xfer;            /* data will be coming */
            icd_cntlr [address].length = 16;            /* accept only the first 16 bytes */
            initiated = TRUE;                           /* log the command initiation */
            break;


        case 0x1F:                                      /* Initiate Self-Test */
            tprintf_bus (da_dev, bus, "Initiate Self-Test\n");

            if_command [address] = initiate_self_test;  /* set up the command */
            if_state [address] = parameter_wait;        /* a parameter must follow */
            icd_cntlr [address].length = 1;             /* set to expect the test ID byte */
            break;


        default:                                        /* an unsupported listen secondary was received */
            abort_command (address, io_program_error,   /* abort and sink any data */
                           error_sink);                 /*   that might accompany the command */
            initiated = TRUE;                           /* log the abort initiation */
            break;
        }
    }

else {                                                      /* otherwise the device is talking */
    da_unit [address].wait = icd_cntlr [address].cmd_time;  /* these are always scheduled and */
    initiated = TRUE;                                       /*   logged as initiated */

    if (if_state [address] == read_wait                 /* if we're waiting for a send data secondary */
      && secondary != 0x00                              /*   but it's not there */
      || if_state [address] == status_wait              /* or a send status secondary, */
      && secondary != 0x08)                             /*   but it's not there */
        abort_command (address, io_program_error,       /*   then abort the pending command */
                       idle);                           /*   and process the new command */

    switch (secondary) {                                /* dispatch the talk secondary */

        case 0x00:                                          /* Send Read Data */
            tprintf_bus (da_dev, bus, "Send Read Data\n");

            if (if_state [address] != read_wait)            /* if we're not expecting it */
                abort_command (address, io_program_error,   /*   then abort and source a data byte */
                               error_source);               /*     tagged with EOI */
            else                                            /* otherwise */
                if_state [address] = command_exec;          /*   the command is ready to execute */
            break;


        case 0x08:                                          /* Read Status */
            tprintf_bus (da_dev, bus, "Send Disc Status\n");

            if (if_state [address] != status_wait)          /* if we're not expecting it, */
                abort_command (address, io_program_error,   /*   abort and source a data byte */
                               error_source);               /*     tagged with EOI */
            else                                            /* otherwise all status commands */
                if_state [address] = read_xfer;             /*   are ready to transfer data */
            break;


        case 0x09:                                      /* CRC (Talk) */
            tprintf_bus (da_dev, bus, "Send CRC\n");

            if_command [address] = crc_talk;            /* set up the command */
            if_state [address] = read_xfer;             /* data will be going */
            break;


        case 0x10:                                          /* Device-Specified Jump */
            tprintf_bus (da_dev, bus, "Return DSJ\n");

            if_command [address] = device_specified_jump;   /* set up the command */
            if_state [address] = read_xfer;                 /* data will be going */
            break;


        case 0x1E:                                      /* Read Loopback */
            tprintf_bus (da_dev, bus, "Read Loopback Record\n");

            if_command [address] = read_loopback;       /* set up the command */
            if_state [address] = read_xfer;             /* data will be going */
            break;


        case 0x1F:                                          /* Return Self-Test Result */
            tprintf_bus (da_dev, bus, "Return Self-Test Result\n");

            if_command [address] = return_self_test_result; /* set up the command */
            if_state [address] = read_xfer;                 /* data will be going */
            icd_cntlr [address].length = 1;                 /* return one byte that indicates */
            buffer [0] = 0;                                 /*   that the self-test passed */
            break;


        default:                                        /* an unsupported talk secondary was received */
            abort_command (address, io_program_error,   /* abort and source a data byte */
                           error_source);               /*   tagged with EOI */
            break;
        }
    }

if (accepted && if_command [address] != amigo_identify) /* if the command was accepted and is not Identify */
    set_response (address, CLEAR);                      /*   then disable the poll response  */

if (da_unit [address].wait > 0)                         /* if service was requested */
    activate_unit (&da_unit [address]);                 /*   then schedule the unit */

if (initiated)                                          /* if a new command was initiated */
    if (if_command [address] == disc_command)
        tprintf (da_dev, TRACE_INCO, "Unit %d position %" T_ADDR_FMT "d %s disc command initiated\n",
                 address, da_unit [address].pos, dl_opcode_name (ICD, icd_cntlr [address].opcode));
    else
        tprintf (da_dev, TRACE_INCO, "Unit %d %s command initiated\n",
                 address, if_command_name [if_command [address]]);

if (if_state [address] != last_state)                   /* if the interface state changed */
    tprintf (da_dev, TRACE_STATE, "Unit %u transitioned from %s state to %s state\n",
             address, if_state_name [last_state], if_state_name [if_state [address]]);

return;
}


/* Process a data byte.

   This routine is called to process a data byte received from the bus for an
   addressed device.  Each device present on the bus that has been addressed to
   listen receives the byte.  Depending on the controller state, the byte
   represents either a command opcode, a command parameter, or a data value.

   On entry, the "address" parameter contains the bus address of the target
   device, which will be a listener.  The "bus" parameter contains the set of
   eight control and eight data lines representing the HP-IB state.

   The routine interprets the data byte in the context of the current device
   controller state, as follows:

     opcode_wait
     -----------
     A Receive Disc Command secondary has been received, and the interface is
     waiting for the opcode that should follow.

     parameter_wait
     --------------
     A disc opcode or interface command has been received, and the interface is
     waiting for a parameter byte that should follow.

     write_xfer
     ----------
     A disc write is in progress, and the interface is waiting for a data byte
     that should follow.

     error_sink
     ----------
     A disc write has terminated with an error, and the interface is waiting to
     absorb all of the remaining data bytes of the transfer.

   Entry in any other state results in an I/O Program Error.
*/

static void process_data (uint32 address, BUS_CABLE bus)
{
const uint8    data       = DATA (bus);                 /* the data value from the bus */
const IF_STATE last_state = if_state [address];         /* the interface state at entry */
t_bool         initiated  = FALSE;

tprintf_bus (da_dev, bus, "Data (device %u accepts)\n", address);

switch (if_state [address]) {                           /* dispatch the interface state */

    case opcode_wait:                                   /* waiting for an opcode */
        buffer [0] = TO_WORD (data, 0);                 /* set the opcode into the buffer */

        if (dl_prepare_command (&icd_cntlr [address],   /* is the command valid? */
                                da_unit, address)) {
            if_state [address] = parameter_wait;        /* set up to get the pad byte */
            icd_cntlr [address].index = 0;              /* reset the word index for the next byte */
            icd_cntlr [address].length =                /* convert the parameter count to bytes */
              icd_cntlr [address].length * 2 + 1;       /*   and include the pad byte */

            tprintf (da_dev, TRACE_CMD, "Unit %u %s command executing\n",
                     address, dl_opcode_name (ICD, icd_cntlr [address].opcode));
            }

        else {                                          /* the disc command is invalid */
            abort_command (address, illegal_opcode,     /* abort the command */
                           error_sink);                 /*   and sink any parameter bytes */
            initiated = TRUE;                           /* log the abort initiation */
            }                                           /* (the unit cannot be busy) */
        break;


    case parameter_wait:                                /* waiting for a parameter */
        put_buffer_byte (&icd_cntlr [address], data);   /* add the byte to the buffer */

        if (icd_cntlr [address].length == 0)            /* is this the last parameter? */
            if (bus & Bus_EOI)                          /* does the host agree? */
                initiated = start_command (address);    /* start the command and log the initiation */

            else {                                          /* the parameter count is wrong */
                abort_command (address, io_program_error,   /* abort the command and sink */
                               error_sink);                 /*   any additional parameter bytes */
                initiated = TRUE;                           /* log the abort initiation */
                }
        break;


    case write_xfer:                                        /* transferring write data */
        if (icd_cntlr [address].length > 0)                 /* if there is more to transfer */
            put_buffer_byte (&icd_cntlr [address], data);   /*   then add the byte to the buffer */

    /* fall through into the error_sink handler */

    case error_sink:                                    /* sinking data after an error */
        if (bus & Bus_EOI)                              /* is this the last byte from the bus? */
            icd_cntlr [address].eod = SET;              /* indicate EOD to the controller */

        di_bus_control (da, address, Bus_NRFD, 0);      /* assert NRFD to hold off the card */

        da_unit [address].wait = icd_cntlr [address].data_time; /* schedule the unit */
        break;


    default:                                            /* data was received in the wrong state */
        abort_command (address, io_program_error,       /* report the error */
                       error_sink);                     /*   and sink any data that follows */
        break;
    }

if (da_unit [address].wait > 0)                         /* if service was requested */
    activate_unit (&da_unit [address]);                 /*   then schedule the unit */

if (initiated)                                          /* if a new command was initiated */
    if (if_command [address] == disc_command)
        tprintf (da_dev, TRACE_INCO, "Unit %d position %" T_ADDR_FMT "d %s disc command initiated\n",
                 address, da_unit [address].pos, dl_opcode_name (ICD, icd_cntlr [address].opcode));
    else
        tprintf (da_dev, TRACE_INCO, "Unit %d %s command initiated\n",
                 address, if_command_name [if_command [address]]);

if (if_state [address] != last_state)                   /* if the interface state changed */
    tprintf (da_dev, TRACE_STATE, "Unit %u transitioned from %s state to %s state\n",
             address, if_state_name [last_state], if_state_name [if_state [address]]);

return;
}


/* Start a command with parameters.

   A command that has been waiting for all of its parameters to be received is
   now ready to start.  If this is a disc command, call the disc library to
   validate the parameters and, if they are OK, to start the command.  Status
   commands return the status values in the sector buffer and the number of
   words that were returned in the buffer length, which we convert to a byte
   count.

   If the disc command was accepted, the library returns a pointer to the unit
   to be scheduled.  For an ICD controller, the unit is always the one currently
   addressed, so we simply test if the return is not NULL.  If it isn't, then we
   set the next interface state as determined by the command that is executing.
   For example, a Read command sets the interface to read_wait status in order
   to wait until the accompanying Send Read Data secondary is received.

   If the return is NULL, then the command was rejected, so we set DSJ = 1 and
   leave the interface state in parameter_wait; the controller status will have
   been set to the reason for the rejection.

   If the next interface state is command_exec, then the disc command is ready
   for execution, and we return TRUE to schedule the unit service.  Otherwise,
   we return FALSE, and the appropriate action will be taken by the caller.

   For all other commands, execution begins as soon as the correct parameters
   are received, so we set command_exec state and return TRUE.  (Only Amigo
   Clear and Initiate Self Test require parameters, so they will be the only
   other commands that must be started here.)


   Implementation notes:

    1. As the ICD implementation does not need to differentiate between unit and
       controller commands, the return value from the dl_start_command routine
       is not used other than as an indication of success or failure.
*/

static t_bool start_command (uint32 unit)
{
if (if_command [unit] == disc_command) {                        /* are we starting a disc command? */
    if (dl_start_command (&icd_cntlr [unit], da_unit, unit)) {  /* start the command; was it successful? */
        icd_cntlr [unit].length = icd_cntlr [unit].length * 2;  /* convert the return length from words to bytes */
        if_state [unit] = next_state [icd_cntlr [unit].opcode]; /* set the next interface state */
        }

    else                                                /* the command was rejected */
        if_dsj [unit] = 1;                              /*   so indicate an error */

    if (if_state [unit] == command_exec)                /* if the command is executing */
        return TRUE;                                    /*   activate the unit */

    else {                                              /* if we must wait */
        da_unit [unit].wait = 0;                        /*   for another secondary, */
        return FALSE;                                   /*   then skip the activation */
        }
    }

else {                                                  /* all other commands */
    if_state [unit] = command_exec;                     /*   execute as soon */
    da_unit [unit].wait = icd_cntlr [unit].cmd_time;    /*     as they */
    return TRUE;                                        /*       are received */
    }
}


/* Abort an in-process command.

   A command sequence partially received via the bus must be aborted.  The cause
   might be an unknown secondary, an illegal disc command opcode, an improper
   secondary sequence (e.g., a Read not followed by Send Read Data), an
   incorrect number of parameters, or unaddressing before the sequence was
   complete.  In any event, the controller and interface are set to an abort
   state, and the DSJ value is set to 1 to indicate an error.
*/

static void abort_command (uint32 unit, CNTLR_STATUS status, IF_STATE state)
{
if_command [unit] = invalid;                            /* indicate an invalid command */
if_state [unit] = state;                                /* set the interface state as directed */
if_dsj [unit] = 1;                                      /* set DSJ to indicate an error condition */

dl_end_command (&icd_cntlr [unit], status);             /* place the disc controller into the wait state */

return;
}


/* Complete an in-process read command.

   An Untalk terminates a Read, Read Full Sector, Read Without Verify, Read With
   Offset, or Cold Load Read command, which must be tied off cleanly by setting
   the end-of-data condition and calling the service routine.  This is required
   only if the read has not already aborted (e.g., for an auto-seek error).

   If a read is in progress, the controller will be busy, and the interface
   state will be either command_exec (if between sectors) or read_xfer (if
   within a sector).  We set up the end phase for the command and schedule the
   disc service to tidy up.

   If a read has aborted, the controller will be waiting, and the interface
   state will be error_source.  In this latter case, we no nothing, as the
   controller has already set the required error status.

   We must be careful NOT to trigger on an Untalk that may follow the opcode and
   precede the Send Read Data sequence.  In this case, the controller will be
   busy, but the interface state will be either read_wait or status_wait.


   Implementation notes:

    1. The test for controller busy is made before calling this routine.  This
       saves the call overhead for the most common case, which is the card is
       being unaddressed after command completion.

    2. There is no need to test if we are processing a disc command, as the
       controller would not be busy otherwise.

    3. If an auto-seek will be needed to continue the read, but the seek will
       fail, then an extra delay is inserted before the service call to start
       the next sector.  Once an Untalk is received, this delay is no longer
       needed, so it is cancelled before rescheduling the service routine.
*/

static void complete_read (uint32 unit)
{
if ((if_state [unit] == command_exec                        /* if a command is executing */
  || if_state [unit] == read_xfer)                          /*   or data is transferring */
  && (dl_classify (icd_cntlr [unit]) == class_read          /* and the controller is executing */
  ||  dl_classify (icd_cntlr [unit]) == class_status)) {    /*   a read or status command */
    icd_cntlr [unit].eod = SET;                             /*     then set the end of data flag */

    if_state [unit] = command_exec;                     /* set to execute */
    da_unit [unit].Phase = end_phase;                   /*   the completion phase */

    sim_cancel (&da_unit [unit]);                       /* cancel the EOT delay */
    da_unit [unit].wait = icd_cntlr [unit].data_time;   /* reschedule for completion */
    }

return;
}


/* Complete an in-process write command.

   Normally, the host sends a byte tagged with EOI to end a Write, Write Full
   Sector, or Initialize command.  However, an Unlisten may terminate a write,
   which must be tied off cleanly by setting the end-of-data condition and
   calling the service routine.  This is required only if the write has not
   already aborted (e.g., for a write-protected disc).

   If a write is in progress, the controller will be busy, and the interface
   state will be either command_exec (if between sectors) or write_xfer (if
   within a sector).  We set up the end phase for the command and schedule the
   disc service to tidy up.

   If a write has aborted, the controller will be waiting, and the interface
   state will be error_sink.  In this latter case, we do nothing, as the
   controller has already set the required error status.

   We must be careful NOT to trigger on the Unlisten that may follow the opcode
   and precede the Receive Write Data sequence.  In this case, the controller
   will be busy, but the interface state will be write_wait.


   Implementation notes:

    1. The test for controller busy is made before calling this routine.  This
       saves the call overhead for the most common case, which is the card is
       being unaddressed after command completion.

    2. There is no need to test if we are processing a disc command, as the
       controller would not be busy otherwise.
*/

static void complete_write (uint32 unit)
{
if ((if_state [unit] == command_exec                    /* if a command is executing */
  || if_state [unit] == write_xfer)                     /*   or data is transferring */
  && dl_classify (icd_cntlr [unit]) == class_write) {   /* and the controller is executing a write */
    icd_cntlr [unit].eod = SET;                         /*   then set the end of data flag */

    if_state [unit] = command_exec;                     /* set to execute */
    da_unit [unit].Phase = end_phase;                   /*   the completion phase */
    da_unit [unit].wait = icd_cntlr [unit].data_time;   /* ensure that the controller will finish */
    }

return;
}


/* Complete an in-process command abort.

   Errors in the command protocol begin an abort sequence that may involve
   sourcing or sinking bytes to allow the sequence to complete as expected by
   the CPU.  Unaddressing the unit terminates the aborted command.

   If an abort is in progress, and the interface is not idle, the end-of-data
   indication is set, and the disc service routine is called directly to process
   the completion of the abort.  The service routine will terminate the
   error_source or error_sink state cleanly and then idle the interface.


   Implementation notes:

    1. The test for an abort-in-progress is made before calling this routine.
       This saves the call overhead for the most common case, which is the card
       is being unaddressed after normal command completion.
*/

static void complete_abort (uint32 unit)
{
if (if_state [unit] != idle) {                          /* is the interface busy? */
    icd_cntlr [unit].eod = SET;                         /* set the end of data flag */
    da_service (&da_unit [unit]);                       /*   and process the abort completion */
    }

return;
}


/* Get a byte from the sector buffer.

   The next available byte in the sector buffer is returned to the caller.  The
   determination of which byte of the 16-bit buffer word to return is made by
   the polarity of the buffer byte count.  The count always begins with an even
   number, as it is set by doubling the word count returned from the disc
   library.  Therefore, because we decrement the count first, the upper byte is
   indicated by an odd count, and the lower byte is indicated by an even count.
   The buffer index is incremented only after the lower byte is returned.
*/

static uint8 get_buffer_byte (CVPTR cvptr)
{
cvptr->length = cvptr->length - 1;                      /* count the byte */

if (cvptr->length & 1)                                  /* is the upper byte next? */
    return UPPER_BYTE (buffer [cvptr->index]);          /* return the byte */
else                                                    /* the lower byte is next */
    return LOWER_BYTE (buffer [cvptr->index++]);        /* return the byte and bump the word index */
}


/* Put a byte into the sector buffer.

   The supplied byte is stored in the sector buffer.  The determination of which
   byte of the 16-bit buffer word to store is made by the polarity of the buffer
   byte count.  The count always begins with an even number, as it is set by
   doubling the word count returned from the disc library.  Therefore, because
   we decrement the count first, the upper byte is indicated by an odd count,
   and the lower byte is indicated by an even count.  The buffer index is
   incremented only after the lower byte is stored.
*/

static void put_buffer_byte (CVPTR cvptr, uint8 data)
{
cvptr->length = cvptr->length - 1;                      /* count the byte */

if (cvptr->length & 1)                                  /* is the upper byte next? */
    buffer [cvptr->index] = TO_WORD (data, 0);          /* save the byte */
else                                                    /* the lower byte is next */
    buffer [cvptr->index++] |= TO_WORD (0, data);       /* merge the byte and bump the word index */
return;
}


/* Activate the unit.

   The specified unit is activated using the unit's "wait" time.  If debugging
   is enabled, the activation is logged to the debug file.
*/

static t_stat activate_unit (UNIT *uptr)
{
t_stat result;

tprintf (da_dev, TRACE_SERV, "Unit %d state %s delay %d service scheduled\n",
         uptr - da_unit, if_state_name [if_state [uptr - da_unit]], uptr->wait);

result = sim_activate (uptr, uptr->wait);               /* activate the unit */
uptr->wait = 0;                                         /* reset the activation time */

return result;                                          /* return the activation status */
}


/* Enable or disable a device's parallel poll response.

   This routine changes the poll response for a specified bus device.  It
   corresponds in hardware to a device's IEEE-488 PPE (parallel poll enable) or
   PPD (parallel poll disable) state.  When the controller conducts a parallel
   poll, the responses of all eight bus devices are placed on the data bus for
   retrieval by the controller.  An enabled poll response indicates to the host
   that a device is requesting service.

   On entry, the "address" parameter is the bus address of the device whose
   poll response is changing, and the "response" parameter indicates whether the
   response should be set (enabled) or cleared (disabled).  The poll response
   bit corresponding to the bus address is included or excluded from the set of
   poll responses.

   If a parallel poll is in progress when a poll response is set, the poll is
   conducted again to reflect the new response.


   Implementation notes:

    1. The set of poll responses is maintained by the channel to improve
       performance, so that we do not need to query each device during a poll to
       ascertain its response.
*/

static void set_response (uint32 address, FLIP_FLOP response)
{
const char *response_name [] = { "disabled", "enabled" };
const BUS_DATA old_response = channel.poll_response;

if (response == SET)                                    /* if the poll response is set */
    channel.poll_response |= PPR (address);             /*   then include the poll bit in the response */
else                                                    /* otherwise we are clearing the poll response */
    channel.poll_response &= ~PPR (address);            /*   so remove the poll bit */

if (channel.poll_response != old_response)
    tprintf (da_dev, TRACE_STATE, "Unit %d parallel poll response %s\n",
             address, response_name [response]);

if (channel.poll_active && response == SET)             /* if a parallel poll is in progress */
    di_bus_control (da, address, BUS_NONE, BUS_NONE);   /*   then conduct it again with the new response */

return;
}
