/* hp2100_lps.c: HP 2100 12653A Line Printer Interface simulator

   Copyright (c) 1993-2016, Robert M. Supnik
   Copyright (c) 2017-2022, J. David Bryan

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
   AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
   ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
   WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the names of the authors shall not be
   used in advertising or otherwise to promote the sale, use or other dealings
   in this Software without prior written authorization from the authors.

   LPS          HP 12653A Line Printer Interface

   11-Mar-22    JDB     Moved execution initialization to "lps_initializer"
   01-Apr-21    JDB     Removed REG_FIT from register definitions (obsolete)
   19-Mar-21    JDB     Converted ftell to position increment for pipes
   09-Jan-21    JDB     Added COMPACT/EXPANDED, TOF, PAPERSTEP, MASTERCLEAR options
                        Added CMD, CSRW, SERV trace options
                        Improved fidelity to interface and printer hardware
   10-Dec-20    JDB     LPS is now enabled by default
   09-Dec-20    JDB     Replaced "io_assert" with "io_assert_ENF"
   28-Nov-20    JDB     Added the "lps_power" routine to handle power state changes
                        Added "lps_set_mode" and "lps_show_mode" to handle modes
                        Removed local SET POWEROFF/POWERON commands
   09-Dec-19    JDB     Removed redundant SIM_SW_REST check in "lps_attach"
   14-Nov-18    JDB     Diagnostic mode now uses GP Register jumper configuration
   26-Jun-18    JDB     Revised I/O model
   03-Aug-17    JDB     Changed perror call for I/O errors to cprintf
   20-Jul-17    JDB     Removed "lps_stopioe" variable and register
   17-Mar-17    JDB     Added "-N" handling to the attach routine
   15-Mar-17    JDB     Changed DEBUG_PRS calls to tprintfs
   13-May-16    JDB     Modified for revised SCP API function parameter types
   10-Feb-12    JDB     Deprecated DEVNO in favor of SC
   28-Mar-11    JDB     Tidied up signal handling
   26-Oct-10    JDB     Changed I/O signal handler for revised signal model
                        Revised detection of CLC at last DMA cycle
   19-Oct-10    JDB     Corrected 12566B (DIAG mode) jumper settings
   26-Jun-08    JDB     Rewrote device I/O to model backplane signals
   10-May-07    RMS     Added UNIT_TEXT flag
   11-Jan-07    JDB     CLC cancels I/O event if DIAG (jumper W9 in "A" pos)
                        Added ioCRS state to I/O decoders
   19-Nov-04    JDB     Added restart when set online, etc.
                        Fixed col count for non-printing chars
   01-Oct-04    JDB     Added SET OFFLINE/ONLINE, POWEROFF/POWERON
                        Fixed status returns for error conditions
                        Fixed handling of non-printing characters
                        Fixed handling of characters after column 80
                        Improved timing model accuracy for RTE
                        Added fast/realistic timing
                        Added debug printouts
   03-Jun-04    RMS     Fixed timing (found by Dave Bryan)
   26-Apr-04    RMS     Fixed SFS x,C and SFC x,C
                        Implemented DMA SRQ (follows FLG)
   25-Apr-03    RMS     Revised for extended file support
   24-Oct-02    RMS     Added microcircuit test features
   30-May-02    RMS     Widened POS to 32b
   03-Dec-01    RMS     Changed DEVNO to use extended SET/SHOW
   07-Sep-01    RMS     Moved function prototypes
   21-Nov-00    RMS     Fixed flag, fbf power up state
                        Added command flop
   15-Oct-00    RMS     Added variable device number support

   References:
     - 2767A Line Printer Operating and Service Manual
         (02767-90002, October 1973)
     - 12653A Line Printer Interface Kit
         (12653-90002, October 1971)
     - General Purpose Register Diagnostic Reference Manual
         (24391-90001, April 1982)
     - HP 12984A Line Printer Subsystem Diagnostic Reference Manual
         (12984-90005, March 1976)


   The HP 12653A Line Printer Interface Kit connects the 2767A printer to the HP
   1000 family.  The subsystem consists of an interface card employing TTL-level
   line drivers and receivers, an interconnecting cable, and an HP 2767A line
   printer.  The interface is supported by RTE and DOS drivers DVR12.  The
   interface supports DMA transfers, but the OS drivers do not use them.

   The 2767A impact printer has a rotating drum with 80 columns of 64 raised
   characters.  ASCII codes 32 through 95 (' ' through '_') form the print
   repertoire.  The printer responds to the control characters FF, LF, and CR
   and prints from 356 to 1110 lines per minute, depending on line length.

   The 80 columns are divided into four zones of 20 characters each that are
   addressed sequentially.  Received characters are buffered in a 20-character
   memory.  When the 20th character is received, the current zone is printed,
   the next zone is addressed, and the memory is reset.  In the absence of print
   command characters, a zone print operation will commence after each group of
   20 characters is transmitted to the printer.  If more than 80 characters are
   sent without an intervening paper advance, zone printing wraps around to zone
   1, and the additional characters overprint those already printed.

   The 2767A does not support a VFU or use an explicit format command.  Instead,
   it reacts to three control characters in the print stream, as follows:

     * CR -- print the characters in the current zone, reset to zone 1, and
             clear the buffer memory.

     * LF -- same as CR, plus advance the paper one line.

     * FF -- same as CR, plus advance the paper to the top of the next form.

   The 12653A outputs seven data bits and a Data Strobe signal to the printer.
   The interface receives Printer Demand and Printer Ready signals from the
   printer.

   STROBE is the positive-true Device Command signal asserted by the Command
   Flip-Flop.  DEMAND is the negative-true Device Flag signal asserted by the
   printer; it denies when the printer is busy and asserts to indicate that the
   printer has completed the requested operation and is ready for the next.
   Printer Ready is the negative-true READY signal asserted by the printer.  It
   asserts to indicate that the printer is operationally ready and no fault
   exists.

   With DEMAND asserted, a programmed STC sets the Command Flip-Flop, asserting
   STROBE.  One microsecond later, DEMAND denies in response, which clears the
   Command Flip-Flop, denying STROBE.  After the printer accepts the character
   on the Data Out lines, DEMAND is asserted again.  For transfers to the
   character buffer, the delay is two microseconds.  If a print cycle results,
   DEMAND remains denied from 0.5 to 35 milliseconds, depending on the specific
   characters in the buffer and the rotational position of the print drum.
   DEMAND assertion sets the Flag Buffer Flip-Flop.  The DEMAND signal is
   reflected in the state of bit 0 of the input register.  Because DEMAND is
   negative-true, bit 0 is set when the printer is busy.

   Printer Ready asserts when power is applied and no fault condition (drum gate
   open, drum speed low, or paper out) exists.  The READY signal is reflected in
   the state of bit 15 of the input register.  Because READY is negative-true,
   bit 15 is set when the printer is not ready.

   At initial power on, READY and DEMAND are both denied.  After the drum comes
   up to speed (which takes about ten seconds), READY asserts.  However, because
   the printer is initially offline, DEMAND remains denied.  Once the printer is
   manually set online, DEMAND asserts.  The resulting status word values, which
   reflect the negative-true printer signals, are:

     Status  Meaning
     ------  --------------------------------------------------------------
     100001  Power off or cable disconnected
     100001  Initial power on, drum speed low
     000001  Power on, drum at speed, paper loaded, printer offline
     000000  Power on, drum at speed, paper loaded, printer online and idle
     000001  Power on, drum at speed, paper loaded, printer online and busy
     100001  Power on, drum at speed, paper out

   The simulator supports a single line printer.  Two output modes are provided:
   an expanded mode that is suitable for retaining printer output as a text
   file, and a compact mode that is suitable for sending the printer output to a
   host-connected physical printer.

   The printer supports realistic and optimized (fast) timing modes.  Realistic
   timing attempts to model the print buffer load, zone print, and paper advance
   operation delays inherent in the physical hardware.  For example, in REALTIME
   mode, output of longer lines takes more time than output of shorter lines,
   and spacing six lines takes approximately six times longer than spacing one
   line.  In FASTTIME mode, all timings are reduced to be "just long enough" to
   satisfy the software driver requirements.

   The line printer diagnostic times print operations, and timing is sensitive
   to the speed of the CPU that is running the diagnostic.  To accommodate this,
   the simulator maintains two sets of real-time delays: one set for the 1000
   E/F machines, and another set for the 21xx and 1000 M machine.  The correct
   set is determined automatically by the CPU type set by the user.

   Attaching the LPS device simulates loading paper into the printer.  Detaching
   simulates removing or running out of paper.

   The printer may be set offline or online.  When a "paper out" condition
   occurs, the printer will print the characters currently in the print buffer
   and then go offline.  Installing paper automatically sets the printer online
   and resumes any output that was suspended.

   A diagnostic mode is provided to simulate the installation of the HP
   1251-0332 diagnostic test (loopback) connector in place of the printer cable.
   This is needed to run the General Purpose Register Diagnostic (DSN 143300)
   on the 12653A interface card.

   A diagnostic mode is provided to simulate the installation of the 1251-0332
   loopback connector, modified to connect pins Z/22 to pins AA/23 as required
   by the General Purpose Register Diagnostic.  This ties the output data lines
   to the input data lines and the Device Command output to the Device Flag
   input.  Entering diagnostic mode also configures the jumpers correctly for
   the diagnostic.

   Diagnostic mode is not used for the 12984A Line Printer Subsystem Diagnostic,
   which tests the printer and interface combination.  Three additional commands
   are provided for use with this diagnostic.  The SET LPS TOF command simulates
   toggling the manual Top of Form switch on the printer.  The SET LPS PAPERSTEP
   command simulates toggling the manual Paper Step switch.  The SET LPS
   MASTERCLEAR command simulates toggling the Master Clear switch on the
   printer's power supply maintenance panel.  The printer also responds to the
   POWER FAIL and POWER RESTORE commands to accommodate the diagnostic.


   The interface responds to I/O instructions as follows:

   Output Data Word format (OTA and OTB):

      15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | -   -   -   -   -   -   -   -   - |      ASCII character      |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+


   Input Data Word format (LIA and LIB):

      15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | N | -   -   -   -   -   -   -   -   -   -   -   -   -   - | B |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     N = Printer Not Ready
     B = Printer Busy
*/



#include "hp2100_defs.h"
#include "hp2100_io.h"



/* Printer program constants.


   Implementation notes:

    1. The size definitions are mixed case to avoid a name clash with PAGE_SIZE
       in <limits.h>.
*/

#define LF                  '\012'              /* line feed */
#define FF                  '\014'              /* form feed */
#define CR                  '\015'              /* carriage return */
#define DEL                 '\177'              /* delete */

#define Zone_Size           20                  /* number of characters per zone */
#define Line_Size           80                  /* number of characters per line */
#define Page_Size           60                  /* number of lines per page */
#define Form_Size           66                  /* number of lines per form */

#define BUFFER_SIZE         (Line_Size + Form_Size * 2) /* characters + max slew * 2 (for CR LF) */

#define LOOPBACK_DELAY      uS (1)              /* diagnostic loopback flag assertion delay */


/* Device flags */

#define DEV_DIAG_SHIFT      (DEV_V_UF + 0)      /* diagnostic VFU tape is installed */
#define DEV_REALTIME_SHIFT  (DEV_V_UF + 1)      /* timing mode is realistic */

#define DEV_DIAG            (1u << DEV_DIAG_SHIFT)
#define DEV_REALTIME        (1u << DEV_REALTIME_SHIFT)


/* Printer unit flags */

#define UNIT_EXPAND_SHIFT   (UNIT_V_UF + 0)     /* printer uses expanded output */
#define UNIT_OFFLINE_SHIFT  (UNIT_V_UF + 1)     /* printer is offline */

#define UNIT_EXPAND         (1u << UNIT_EXPAND_SHIFT)
#define UNIT_OFFLINE        (1u << UNIT_OFFLINE_SHIFT)
#define UNIT_ONLINE         0


/* Simulator modes */

typedef enum {                                  /* Simulator modes */
    Fast_Time = 0,                              /*   use optimized timing (as an array index) */
    Real_Time = 1,                              /*   use realistic timing (as as array index) */
    Printer,                                    /*   connect the interface to the printer */
    Diagnostic,                                 /*   connect the interface to the loopback connector */
    TOF,                                        /*   do a manual top-of-form */
    Step,                                       /*   do a manual paper step */
    Clear                                       /*   do a manual master clear */
    } SIM_MODES;


/* Delay properties table.

   The delay properties table contains entries for the optimized and realistic
   timing specifications for the supported printer.  The times represent the
   delays for mechanical and electronic operations.  Delay values are in event
   tick counts; macros are used to convert from times to ticks.

   Element 0 represents the FASTTIME delays.  Elements 1 and 2 represent
   REALTIME delays for 1000 E/F and 21xx/1000 M CPUs, respectively.  The
   separate entries are necessary as the diagnostic uses two different
   mechanisms to determine if the printer is operating with specification.

   The line printer diagnostic checks completion times, so realistic timing must
   be used.  Because simulator timing is in instructions, and because the
   diagnostic uses the TIMER instruction (~1580 executions per millisecond) when
   running on a 1000-E/F but a software timing loop (~400-600 executions per
   millisecond) when running on anything else, realistic timings are decreased
   by three-fourths when not executing on an E/F.


   Implementation notes:

    1. The printer acknowledges before the print motion has stopped to allow for
       continuous slew, so the supplied times are a bit less than the calculated
       operation times from the manual.
*/

typedef struct {
    int32  buffer_load;                         /* per-character transfer time */
    int32  zone_print;                          /* per-zone printing time */
    int32  advance;                             /* paper advance time per line */
    int32  power_delay;                         /* power-on ready delay time */
    } DELAY_PROPS;

static const DELAY_PROPS fast_times = {         /* initial fast timing delays */
   /*  buffer     zone       paper    power  */
   /*   load      print     advance   delay  */
   /* --------  ---------  ---------  ------ */
       uS (1),  uS (500),  uS (500),  mS (1)
    };

static DELAY_PROPS times [] = {                 /* timing delays, indexed by SIM_MODE and CPU type */
   /*  buffer     zone       paper     power   */
   /*   load      print     advance    delay   */
   /* --------  ---------  ---------  -------- */
    {  uS (1),  uS (500),  uS (500),  mS (1)   },       /* Fast_Time timing */
    {  uS (2),  mS (35),   mS (11),   mS (100) },       /* Real_Time timing on 1000 E/F CPU */
    {  uS (1),  mS (8.75), mS (2.75), mS (25)  }        /* Real_Time timing on 21xx/1000 M CPU */
    };


/* Printer control word.

      15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | -   -   -   -   -   -   -   -   - |      ASCII character      |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
*/

#define DATA_MASK           0177u               /* printer uses only 7 bits for data */


/* Printer status word.

      15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | N | -   -   -   -   -   -   -   -   -   -   -   -   -   - | B |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     N = Not Ready
     B = Busy
*/

#define ST_NOT_READY        0100000u            /* not ready status */
#define ST_BUSY             0000001u            /* busy status */

#define ST_POWEROFF         (ST_BUSY | ST_NOT_READY)

static const BITSET_NAME prt_status_names [] = {        /* Printer status word names */
    "\1not ready\0ready",                               /*   bit 15 */
    NULL,                                               /*   bit 14 */
    NULL,                                               /*   bit 13 */
    NULL,                                               /*   bit 12 */
    NULL,                                               /*   bit 11 */
    NULL,                                               /*   bit 10 */
    NULL,                                               /*   bit  9 */
    NULL,                                               /*   bit  8 */
    NULL,                                               /*   bit  7 */
    NULL,                                               /*   bit  6 */
    NULL,                                               /*   bit  5 */
    NULL,                                               /*   bit  4 */
    NULL,                                               /*   bit  3 */
    NULL,                                               /*   bit  2 */
    NULL,                                               /*   bit  1 */
    "\1busy\0idle"                                      /*   bit  0 */
    };

static const BITSET_FORMAT prt_status_format =          /* names, offset, direction, alternates, bar */
    { FMT_INIT (prt_status_names, 0, msb_first, has_alt, no_bar) };


/* Interface state */

typedef struct {
    HP_WORD      output_word;                   /* output data register */
    HP_WORD      status_word;                   /* input data register */
    FLIP_FLOP    command;                       /* command flip-flop */
    FLIP_FLOP    control;                       /* control flip-flop */
    FLIP_FLOP    flag;                          /* flag flip-flop */
    FLIP_FLOP    flag_buffer;                   /* flag buffer flip-flop */
    t_bool       strobe;                        /* STROBE signal to the printer */
    t_bool       demand;                        /* DEMAND signal from the printer */
    } CARD_STATE;

static CARD_STATE lps;                          /* per-card state */


/* Printer state */

static POWER_STATE power_state     = power_on;  /* the printer power state */
static uint8       overprint_char  = DEL;       /* character to use if overprinted */
static uint32      current_line    = 1;         /* current form line */
static uint32      buffer_index    = 0;         /* current index into the print buffer */
static uint32      overprint_index = 0;         /* the "high-water" mark while overprinting */
static uint8       buffer [BUFFER_SIZE];        /* character buffer */

static const DELAY_PROPS *dlyptr = &times [Fast_Time];   /* pointer to the event delay times to use */


/* Interface local SCP support routines */

static IO_INTERFACE   lps_interface;
static IO_INITIALIZER lps_initializer;
static IO_POWER       lps_power;

static t_stat lps_service (UNIT   *uptr);
static t_stat lps_reset   (DEVICE *dptr);


/* Interface local utility routines */

static void activate_unit (UNIT *uptr);
static void report_error  (FILE *stream);


/* Printer local SCP support routines */

static t_stat lps_attach    (UNIT *uptr, char *cptr);
static t_stat lps_detach    (UNIT *uptr);
static t_stat lps_set_xline (UNIT *uptr, int32 value, char *cptr, void *desc);
static t_stat lps_set_mode  (UNIT *uptr, int32 value, char  *cptr, void *desc);
static t_stat lps_show_mode (FILE *st,   UNIT  *uptr, int32 value, void *desc);


/* Interface SCP data structures */


/* Unit list */

#define UNIT_FLAGS          (UNIT_ATTABLE | UNIT_SEQ | UNIT_EXPAND | UNIT_OFFLINE)

static UNIT lps_unit [] = {
    { UDATA (&lps_service, UNIT_FLAGS, 0), 0 }
    };


/* Device information block */

static DIB lps_dib = {
    lps_interface,                              /* the device's I/O interface function pointer */
    lps_initializer,                            /* the execution initialization function pointer */
    lps_power,                                  /* the power state function pointer */
    LPS,                                        /* the device's select code (02-77) */
    0,                                          /* the card index */
    "12653A Line Printer Interface",            /* the card description */
    NULL                                        /* the ROM description */
    };


/* Register list */

static REG lps_reg [] = {
/*    Macro   Name    Location                       Radix    Width    Offset      Depth          Flags       */
/*    ------  ------  -----------------------------  -----  ---------  ------  ------------  ---------------- */
    { FLDATA (CMD,    lps.command,                                        0)                                  },
    { FLDATA (CTL,    lps.control,                                        0)                                  },
    { FLDATA (FLG,    lps.flag,                                           0)                                  },
    { FLDATA (FBF,    lps.flag_buffer,                                    0)                                  },

    { FLDATA (DEMAND, lps.demand,                                         0)                                  },
    { FLDATA (STROBE, lps.strobe,                                         0)                                  },

    { ORDATA (OUTPUT, lps.output_word,                         16),                          PV_RZRO | REG_X  },
    { ORDATA (STATUS, lps.status_word,                         16),                          PV_RZRO          },

    { DRDATA (PRLINE, current_line,                             8),                          PV_LEFT          },
    { DRDATA (BUFIDX, buffer_index,                             8),                          PV_LEFT          },
    { BRDATA (PRTBUF, buffer,                         8,        8,             BUFFER_SIZE), PV_RZRO | REG_A  },
    { DRDATA (OVPIDX, overprint_index,                          8),                          PV_LEFT          },
    { ORDATA (OVPCHR, overprint_char,                           8),                          PV_RZRO | REG_A  },
    { DRDATA (POWER,  power_state,                              2),                          REG_RO           },
    { DRDATA (POS,    lps_unit [0].pos,                     T_ADDR_W),                       PV_LEFT          },

    { DRDATA (BTIME,  times [Fast_Time].buffer_load,           24),                          PV_LEFT | REG_NZ },
    { DRDATA (PTIME,  times [Fast_Time].zone_print,            24),                          PV_LEFT | REG_NZ },
    { DRDATA (STIME,  times [Fast_Time].advance,               24),                          PV_LEFT | REG_NZ },
    { DRDATA (DTIME,  times [Fast_Time].power_delay,           24),                          PV_LEFT | REG_NZ },

      DIB_REGS (lps_dib),

    { NULL }
    };


/* Modifier list */

static MTAB lps_mod [] = {
/*    Mask Value     Match Value    Print String       Match String  Validation      Display  Descriptor */
/*    -------------  -------------  -----------------  ------------  --------------  -------  ---------- */
    { UNIT_OFFLINE,  UNIT_OFFLINE,  "offline",         "OFFLINE",    &lps_set_xline, NULL,    NULL       },
    { UNIT_OFFLINE,  0,             "online",          "ONLINE",     &lps_set_xline, NULL,    NULL       },

    { UNIT_EXPAND,   UNIT_EXPAND,   "expanded output", "EXPAND",     NULL,           NULL,    NULL       },
    { UNIT_EXPAND,   0,             "compact output",  "COMPACT",    NULL,           NULL,    NULL,      },

/*    Entry Flags          Value       Print String  Match String   Validation     Display         Descriptor        */
/*    -------------------  ----------  ------------  -------------  -------------  --------------  ----------------- */
    { MTAB_XDV,            Fast_Time,  NULL,         "FASTTIME",    &lps_set_mode, NULL,           NULL              },
    { MTAB_XDV,            Real_Time,  NULL,         "REALTIME",    &lps_set_mode, NULL,           NULL              },
    { MTAB_XDV,            Printer,    NULL,         "PRINTER",     &lps_set_mode, NULL,           NULL              },
    { MTAB_XDV,            Diagnostic, NULL,         "DIAGNOSTIC",  &lps_set_mode, NULL,           NULL              },
    { MTAB_XDV,            TOF,        NULL,         "TOF",         &lps_set_mode, NULL,           NULL              },
    { MTAB_XDV,            Step,       NULL,         "PAPERSTEP",   &lps_set_mode, NULL,           NULL              },
    { MTAB_XDV,            Clear,      NULL,         "MASTERCLEAR", &lps_set_mode, NULL,           NULL              },

    { MTAB_XDV,            0,          "MODES",      NULL,          NULL,          &lps_show_mode, NULL              },

    { MTAB_XDV,            1u,         "SC",         "SC",          &hp_set_dib,   &hp_show_dib,   (void *) &lps_dib },
    { MTAB_XDV | MTAB_NMO, ~1u,        "DEVNO",      "DEVNO",       &hp_set_dib,   &hp_show_dib,   (void *) &lps_dib },
    { 0 }
    };


/* Debugging trace list */

static DEBTAB lps_deb [] = {
    { "CMD",   TRACE_CMD   },                   /* trace interface or controller commands */
    { "CSRW",  TRACE_CSRW  },                   /* trace interface control, status, read, and write actions */
    { "STATE", TRACE_STATE },                   /* trace state changes */
    { "SERV",  TRACE_SERV  },                   /* trace unit service scheduling calls and entries */
    { "XFER",  TRACE_XFER  },                   /* trace data transmissions */
    { "IOBUS", TRACE_IOBUS },                   /* interface I/O bus signals and data words */
    { NULL,    0           }
    };


/* Device descriptor */

DEVICE lps_dev = {
    "LPS",                                      /* device name */
    lps_unit,                                   /* unit array */
    lps_reg,                                    /* register array */
    lps_mod,                                    /* modifier array */
    1,                                          /* number of units */
    10,                                         /* address radix */
    31,                                         /* address width */
    1,                                          /* address increment */
    8,                                          /* data radix */
    8,                                          /* data width */
    NULL,                                       /* examine routine */
    NULL,                                       /* deposit routine */
    &lps_reset,                                 /* reset routine */
    NULL,                                       /* boot routine */
    &lps_attach,                                /* attach routine */
    &lps_detach,                                /* detach routine */
    &lps_dib,                                   /* device information block */
    DEV_DISABLE | DEV_DEBUG,                    /* device flags */
    0,                                          /* debug control flags */
    lps_deb,                                    /* debug flag name table */
    NULL,                                       /* memory size change routine */
    NULL                                        /* logical device name */
    };



/* Interface local SCP support routines */



/* Line printer interface.

   The line printer interface is installed on the I/O bus and receives I/O
   commands from the CPU and DMA/DCPC channels.  In simulation, the asserted
   signals on the bus are represented as bits in the inbound signal set.  Each
   signal is processed sequentially in numerical order.


   Implementation notes:

    1. In hardware, an STC signal sets both the Control flip-flop and the
       Command flip-flop.  The latter asserts STROBE to the printer, which
       denies DEMAND in response.  The trailing edge of DEMAND clears the
       Command flip-flop and denies STROBE.  When the printer completes its
       operation, it reasserts DEMAND to complete the handshake.

       In simulation, two service routine calls are scheduled -- the first after
       STROBE asserts to perform the indicated operation, and the second to
       reassert DEMAND after the operation completes.

    2. The interface does not inhibit STROBE if the printer is offline.  STROBE
       will assert, but then nothing will happen until DEMAND asserts.  Then,
       with STROBE already asserted, the printer will deny DEMAND, and the
       normal handshake sequence will continue.  In simulation, we skip the
       first event service call if DEMAND is denied and depend on the SET ONLINE
       routine to reschedule the event service is STROBE is asserted.

    3. The timing delays selected depend on the current CPU model.  Because we
       do not get notification of CPU model changes, we can either check the CPU
       model when scheduling each service call, or we can process the PON
       assertion that occurs (with ENF and SIR) when the I/O system is
       initialized prior to the resumption of execution.  As the former occurs
       twice for each character output, we elect to set up the timing delay
       pointer at the latter occurrence, which occurs only once per run
       invocation.  PON is also asserted during an operator PRESET, but it does
       no harm to set up the pointers there as well, so we don't bother to
       qualify the signal further.
*/

static SIGNALS_VALUE lps_interface (const DIB *dibptr, INBOUND_SET inbound_signals, HP_WORD inbound_value)
{
INBOUND_SIGNAL signal;
INBOUND_SET    working_set = inbound_signals;
SIGNALS_VALUE  outbound    = { ioNONE, 0 };
t_bool         irq_enabled = FALSE;

while (working_set) {                                   /* while signals remain */
    signal = IONEXTSIG (working_set);                   /*   isolate the next signal */

    switch (signal) {                                   /* dispatch the I/O signal */

        case ioCLF:                                     /* Clear Flag flip-flop */
            lps.flag_buffer = CLEAR;                    /* reset the flag buffer */
            lps.flag        = CLEAR;                    /*   and flag flip-flops */
            break;


        case ioSTF:                                     /* Set Flag flip-flop */
            lps.flag_buffer = SET;                      /* set the flag buffer flip-flop */
            break;


        case ioENF:                                     /* Enable Flag */
            if (lps.flag_buffer == SET)                 /* if the flag buffer flip-flop is set */
                lps.flag = SET;                         /*   then set the flag flip-flop */
            break;


        case ioSFC:                                     /* Skip if Flag is Clear */
            if (lps.flag == CLEAR)                      /* if the flag flip-flop is clear */
                outbound.signals |= ioSKF;              /*   then assert the Skip on Flag signal */
            break;


        case ioSFS:                                     /* Skip if Flag is Set */
            if (lps.flag == SET)                        /* if the flag flip-flop is set */
                outbound.signals |= ioSKF;              /*   then assert the Skip on Flag signal */
            break;


        case ioIOI:                                     /* I/O Data Input */
            outbound.value = lps.status_word;           /* return the current printer status */

            tprintf (lps_dev, TRACE_CSRW, "Status is %s\n",
                     fmt_bitset (outbound.value, prt_status_format));
            break;


        case ioIOO:                                     /* I/O Data Output */
            lps.output_word = inbound_value;            /* save the character or format word */

            tprintf (lps_dev, TRACE_CSRW, "Control is character %s\n",
                     fmt_char (inbound_value & DATA_MASK));
            break;


        case ioPOPIO:                                   /* Power-On Preset to I/O */
            lps.flag_buffer = SET;                      /* set the flag buffer flip-flop */
            lps.output_word = 0;                           /*   and clear the output register */
            break;


        case ioCRS:                                     /* Control Reset */
            lps.control = CLEAR;                        /* clear the control flip-flop */
            lps.command = CLEAR;                        /*   and the command flip-flop */
            lps.strobe = FALSE;                         /*     and deny STROBE to the printer */

            sim_cancel (lps_unit);                     /*   and cancel any printing in progress */
            break;


        case ioCLC:                                     /* Clear Control flip-flop */
            lps.control = CLEAR;                        /* clear the control flip-flop */
            lps.command = CLEAR;                        /*   and the command flip-flop */
            lps.strobe = FALSE;                         /*     and deny STROBE to the printer */
            break;


        case ioSTC:                                     /* Set Control flip-flop */
            lps.control = SET;                          /* set the control flip-flop */
            lps.command = SET;                          /*   and the command flip-flop */

            if (lps_dev.flags & DEV_DIAG) {             /* if the card is in diagnostic mode */
                lps.status_word = lps.output_word;      /*   then loop the data back */

                tprintf (lps_dev, TRACE_XFER, "Output data word %06o looped back to input\n",
                         lps.output_word);

                activate_unit (lps_unit);               /* schedule the service routine */
                }

            else {                                      /* otherwise the printer is connected */
                lps.strobe = TRUE;                      /*   so assert STROBE to the printer */

                if (lps.demand == TRUE) {               /* if the printer is ready */
                    lps_unit [0].wait = 0;              /*   then set for immediate service entry */
                    activate_unit (lps_unit);           /*     and schedule the service routine */
                    }
                }
            break;


        case ioSIR:                                     /* Set Interrupt Request */
            if (lps.control & lps.flag)                 /* if the control and flag flip-flops are set */
                outbound.signals |= cnVALID;            /*   then deny PRL */
            else                                        /* otherwise */
                outbound.signals |= cnPRL | cnVALID;    /*   conditionally assert PRL */

            if (lps.control & lps.flag & lps.flag_buffer)   /* if the control, flag, and flag buffer flip-flops are set */
                outbound.signals |= cnIRQ;                  /*   then conditionally assert IRQ */

            if (lps.flag == SET)                        /* if the flag flip-flop is set */
                outbound.signals |= ioSRQ;              /*   then assert SRQ */
            break;


        case ioIAK:                                     /* Interrupt Acknowledge */
            lps.flag_buffer = CLEAR;                    /* clear the flag buffer flip-flop */
            break;


        case ioIEN:                                     /* Interrupt Enable */
            irq_enabled = TRUE;                         /* permit IRQ to be asserted */
            break;


        case ioPRH:                                         /* Priority High */
            if (irq_enabled && outbound.signals & cnIRQ)    /* if IRQ is enabled and conditionally asserted */
                outbound.signals |= ioIRQ | ioFLG;          /*   then assert IRQ and FLG */

            if (not irq_enabled || outbound.signals & cnPRL)    /* if IRQ is disabled or PRL is conditionally asserted */
                outbound.signals |= ioPRL;                      /*   then assert it unconditionally */
            break;


        case ioEDT:                                     /* not used by this interface */
        case ioPON:                                     /* not used by this interface */
            break;
        }

    IOCLEARSIG (working_set, signal);                   /* remove the current signal from the set */
    }                                                   /*   and continue until all signals are processed */

return outbound;                                        /* return the outbound signals and value */
}


/* Line printer initializer.

   This routine is called from the instruction execution prelude to set up the
   internal state of the line printer controller prior to instruction execution.

   The routine sets up the timing delays to be used.  The delays are dependent
   on the REALTIME vs. FASTTIME setting, as well as the CPU model.
*/

static t_bool lps_initializer (DEVICE *dptr, DIB *dibptr)
{
if (dptr->flags & DEV_REALTIME)             /* if the printer is in real-time mode */
    dlyptr = &times [Fast_Time];            /*   then point at the fast times */
else if (cpu_configuration & CPU_1000_E_F)  /* otherwise if this is a 1000 E/F CPU */
    dlyptr = &times [Real_Time];            /*   the point at the E/F realistic times */
else                                        /* otherwise */
    dlyptr = &times [Real_Time + 1];        /*   point at the 21xx/1000 M CPU times */

return TRUE;
}


/* Set the new line printer power state.

   This routine is called by the POWER command to set the printer into a new
   power state indicated by the "new_state" parameter.  The "uptr" parameter is
   not used.

   When power is removed from the 2767, the printer immediately returns
   power-off status.  Applying power will initially return power-off status, but
   this will change to power-on status within ten seconds.

   In simulation, receiving either power_failing or power_off will set the
   printer to the power_off state.  Receiving power_returning will schedule the
   printer unit for the power-on delay; receiving power_on will bypass the
   delay.  When the event timer expires, the printer will be set to the power_on
   state.
*/

static t_stat lps_power (UNIT *uptr, POWER_STATE new_state)
{
if (new_state == power_failing)                         /* power failing is treated */
    new_state = power_off;                              /*   identically to power off */

if (new_state == power_state                            /* if the new state is the same as the current state */
  || lps_dev.flags & DEV_DIAG)                          /*   or the card is in diagnostic mode */
    return SCPE_NOFNC;                                  /*     then the POWER command is not allowed */

else {                                                  /* otherwise */
    power_state = new_state;                            /*   establish the new state */

    if (new_state == power_off) {                       /* if the printer power is off */
        lps.status_word = ST_POWEROFF;                  /*   then set the power-off status */
        lps_unit [0].flags |= UNIT_OFFLINE;             /*     and set the printer offline */

        sim_cancel (lps_unit);                          /* cancel any pending print operation */
        }

    else if (new_state == power_returning) {            /* if power is returning */
        lps_unit [0].wait = dlyptr->power_delay;        /*   then schedule the delay */
        activate_unit (lps_unit);                       /*      until the drum is up to speed */
        }

    tprintf (lps_dev, TRACE_STATE, "Power is %s\n",
             hp_power_names [new_state]);

    return SCPE_OK;                                     /* the command completed successfully */
    }
}


/* Service the printer.

   The printer transfer service is called to output a character to the printer
   buffer or to output a format command that causes the buffered line to be
   printed with specified paper movement.

   In hardware, the interface places a character or format code on the lower
   seven data out lines and asserts STROBE to the printer.  The printer responds
   by denying DEMAND.  The interface then denies STROBE and waits for the
   printer to reassert DEMAND to indicate that the buffer load or print
   operation is complete.

   In simulation, this service routine is called twice for each transfer.  It is
   called immediately when STROBE is asserted with DEMAND asserted and then
   after a variable delay with STROBE and DEMAND denied.  In response to the
   former call, the routine denies DEMAND and STROBE, loads the character buffer
   or prints the buffered line, and then sets up an event delay corresponding to
   the operation performed.  In response to the latter call, the routine asserts
   DEMAND and sets the flag buffer flip-flop.

   Control character LF, FF, and CR are interpreted as printer format commands.
   All other characters are interpreted as data to be loaded into the print
   buffer.  Only characters in the octal range 40-137 (' ' through '_') are
   printable.  If a character is sent but the printer cannot print it, a space
   is loaded in its place.

   The printer can buffer up to 20 characters before a "zone print" operation is
   performed automatically.  The simulator models this action as a longer event
   delay when printing occurs.  If more than 80 characters are sent without an
   intervening format command, printing wraps around and begins to overprint the
   current line.

   A format command prints the current buffer, resets the buffer pointer to
   column 1, and then advances the paper by the prescribed amount (one line for
   LF, to the top of the next form for FF, or not at all for CR).  Two output
   modes are provided: compact and expanded.  These modes affect the way the
   attached printer output file is written.

   In compact mode, a printed line is terminated by a CR LF pair, but subsequent
   line spacing is performed by LFs alone.  Also, a top-of-form request will
   emit a FF character instead of the number of LFs required to reach the top of
   the next form, and overprinting is handled by emitting a lone CR at the end
   of the line.  This mode is used when the printer output file will be sent to
   a physical printer connected to the host.

   In expanded mode, paper advance is handled solely by emitting CR LF pairs.
   Overprinting is handled by merging characters in the buffer.  This mode is
   used where the printer output file will be saved or manipulated as a text
   file.

   The printer automatically skips the page perforation after the 60th line is
   printed on a given page.  Six lines -- three before and three after the
   perforation -- are skipped.

   If the stream write fails, an error message is displayed on the simulation
   console, the printer is set offline, and SCPE_IOERR is returned to cause a
   simulation stop to give the user the opportunity to fix the problem.
   Simulation may then be resumed, either with the printer set back online if
   the problem is fixed, or with the printer remaining offline if the problem is
   uncorrectable.

   In diagnostic mode, DEVICE COMMAND is looped back to DEVICE FLAG, so service
   entry only sets the Flag Buffer flip-flop and returns.

   Entry also occurs when power is returning to the printer.  When power is
   applied, the printer takes several seconds to bring the print drum up to
   speed before asserting the PRINTER READY signal.  This delay is implemented
   by scheduling a service entry in the "power returning" state.  On entry, the
   print buffer is cleared, but the paper position is retained.  The printer
   must be set online manually before printing may be resumed.


   Implementation notes:

    1. A printer going offline, either via a user action or by a paper-out
       condition, does not complete the last handshake.  Instead, DEMAND remains
       denied, so the interface flag never sets.  Typically, this causes an I/O
       timeout in the OS driver, which issues a CLC and downs the device.  When
       the printer is set back online, DEMAND asserts, which sets the flag.
       However, the CLC has cleared the control flip-flop, so no interrupt is
       generated.  When the device is upped, the incomplete request is reissued;
       this results in either a duplicate printed line or a duplicate paper
       movement.

    2. Because attached files are opened in binary mode, newline translation
       (i.e., from LF to CR LF) is not performed by the host system.  Therefore,
       we write explicit CR LF pairs to end lines, even in compact mode, as
       required for fidelity to HP peripherals.  If bare LFs are used by the
       host system, the printer output file must be postprocessed to remove the
       CRs.

    3. Overprinting in expanded mode is simulated by merging the lines in the
       buffer.  A format command to suppress spacing resets the buffer index but
       saves the previous buffer length as a "high water mark" that will be
       extended if the overlaying line is longer.  This process may be repeated
       as many times as desired before issuing a format command that prints the
       buffer.

       When overlaying characters, if a space overlays a printing character, a
       printing character overlays a space, or a printing character overlays
       itself, then the printing character is retained.  Otherwise, an
       "overprint character" (which defaults to DEL, but can be changed by the
       user) replaces the character in the buffer.

    4. As a convenience to the user, the printer output stream is flushed when a
       TOF operation is performed.  This permits inspection of the output file
       from an independent process while output is ongoing.

    5. This routine will not be entered with the printer unit unattached.
       DEMAND must be asserted, i.e., the printer must be online (and therefore
       attached) before an STC will schedule the service, and "lps_detach" will
       cancel the service event).  So explicit protection against "fileref"
       being NULL is not required.
*/

static t_stat lps_service (UNIT *uptr)
{
uint8  data_byte;
uint32 slew_count;

tprintf (lps_dev, TRACE_SERV, "Printer service entered\n");

if (lps_dev.flags & DEV_DIAG) {                         /* if the card is in diagnostic mode */
    lps.command = CLEAR;                                /*   then clear the command flip-flop */

    lps.flag_buffer = SET;                              /* set the flag buffer */
    io_assert_ENF (&lps_dib);                           /*   and flag flip-flops */
    }

else if (power_state == power_returning) {              /* otherwise if the drum is back up to speed */
    power_state = power_on;                             /*   then change the state */

    buffer_index = 0;                                   /* the buffer is lost during a power outage */
    overprint_index = 0;                                /*   so clear the overprint index too */

    if (uptr->flags & UNIT_ATT)                         /* if paper is loaded */
        lps.status_word &= ~ST_NOT_READY;               /*   then the printer is busy until put online */

    tprintf (lps_dev, TRACE_STATE, "Power is %s\n",
             hp_power_names [power_on]);
    }

else if (not lps.demand) {                              /* otherwise if DEMAND is denied */
    lps.demand = TRUE;                                  /*   then assert it to complete the handshake */

    lps.flag_buffer = SET;                              /* DEMAND assertion sets */
    io_assert_ENF (&lps_dib);                           /*   the flag buffer and flag flip-flops */
    }

else if (lps.strobe) {                                  /* otherwise if STROBE asserts with DEMAND asserted */
    lps.demand  = FALSE;                                /*   then deny DEMAND */
    lps.command = CLEAR;                                /*     which clears the command flip-flop */
    lps.strobe  = FALSE;                                /*       which denies STROBE */

    uptr->wait = dlyptr->buffer_load;                   /* schedule the buffer load delay */

    data_byte = (uint8) (lps.output_word & DATA_MASK);  /* only the lower 7 bits are sent to the printer */

    if (data_byte == LF || data_byte == FF || data_byte == CR) {    /* if printing is requested */
        tprintf (lps_dev, TRACE_XFER, "Print command %s sent to printer\n",
                 fmt_char (data_byte));

        if (buffer_index > 0)                           /* if characters will be printed */
            uptr->wait += dlyptr->zone_print;           /*   then include the zone printing delay */

        if (overprint_index > buffer_index)             /* if the overprinted line is longer than the current line */
            buffer_index = overprint_index;             /*   then extend the current buffer index */

        if (data_byte == CR                             /* if this is a "print without advancing" request */
          && uptr->flags & UNIT_EXPAND) {               /*   and the printer is in expanded mode */
            if (buffer_index > overprint_index)         /*     then if beyond the current overprinted length */
                overprint_index = buffer_index;         /*       then extend the overprinted line */

            buffer_index = 0;                           /* reset the buffer index to overprint the next character */
            }

        else {                                          /* otherwise the buffer will be printed */
            if (buffer_index > 0) {                     /*   so if characters are present, then trace it */
                tprintf (lps_dev, TRACE_CMD, "Printed %u character%s on line %u\n",
                         buffer_index, (buffer_index == 1 ? "" : "s"), current_line);

                buffer [buffer_index++] = CR;           /* tie off the buffer with a CR */

                if (data_byte != CR) {                  /* if paper advance will occur */
                    buffer [buffer_index++] = LF;       /*   then include a LF too */
                    current_line++;                     /*     and increment the line counter */
                    }
                }

            else if (data_byte == LF) {                 /* otherwise if this is a bare LF */
                tprintf (lps_dev, TRACE_CMD, "Printer advanced 1 line from line %u\n",
                         current_line);

                if (uptr->flags & UNIT_EXPAND)          /*   then if the printer is in expanded mode */
                    buffer [buffer_index++] = CR;       /*      then precede it with a CR */

                buffer [buffer_index++] = LF;           /* add the LF to the buffer */
                current_line++;                         /*   and bump the line counter */
                }

            else if (data_byte == CR)                   /* otherwise if this is a bare CR */
                buffer [buffer_index++] = CR;           /*   then place it in the buffer for compact mode */

            if (data_byte == FF                             /* if an explicit form feed is requested */
              || current_line > Page_Size) {                /*   or printing has advanced into the perf region */
                slew_count = Form_Size - current_line + 1;  /*     then get the number of lines to slew to TOF */

                uptr->wait += slew_count * dlyptr->advance; /* account the paper advance delay */

                tprintf (lps_dev, TRACE_CMD, "Printer advanced %u lines from line %u to line 1\n",
                         slew_count, current_line);

                if (uptr->flags & UNIT_EXPAND)          /* if the printer is in expanded mode */
                    while (slew_count-- > 0) {          /*   then while movement is needed */
                        buffer [buffer_index++] = CR;   /*     send a CR */
                        buffer [buffer_index++] = LF;   /*       and LF to the buffer */
                        }

                else                                    /* otherwise in compact mode */
                    buffer [buffer_index++] = FF;       /*   so emit a FF to move to the TOF */

                current_line = 1;                       /* reset line to the TOF */
                }

            uptr->pos += sizeof buffer [0]                  /* write the buffer to the printer file */
              * (t_addr) fwrite (buffer, sizeof buffer [0], /*   and update the file position */
                                 buffer_index, uptr->fileref);

            buffer_index = 0;                           /* reset the buffer index */
            overprint_index = 0;                        /*   and any existing overprint index */

            if (current_line == 1)                      /* if a TOF request was performed */
                fflush (uptr->fileref);                 /*   then flush the file buffer for inspection */

            if (ferror (uptr->fileref)) {               /* if a host file system error occurred */
                report_error (uptr->fileref);           /*   then report the error to the console */

                lps_set_xline (uptr, UNIT_OFFLINE, NULL, NULL); /* set the printer offline */
                return SCPE_IOERR;                              /*   and stop the simulator */
                }
            }
        }

    else {                                              /* otherwise this is a not a printer command */
        if (data_byte < ' ' || data_byte > '_') {       /*   so if the character is non-printing */
            tprintf (lps_dev, TRACE_XFER, "Character %s erased\n",
                     fmt_char (data_byte));

            data_byte = ' ';                            /*     then substitute a space */
            }

        if (buffer_index == Line_Size) {                /* if the buffer is full */
            overprint_index = buffer_index;             /*   then this character will overprint */

            buffer_index = 0;                           /* reset the index to the buffer start */

            tprintf (lps_dev, TRACE_CMD, "Buffer overflow caused overprinting on line %u\n",
                     current_line);
            }

        if (overprint_index == 0                        /* if not overprinting */
          || buffer_index >= overprint_index            /*   or past the current buffer limit */
          || buffer [buffer_index] == ' ')              /*   or overprinting a blank */
            buffer [buffer_index] = data_byte;          /*     then store the character */

        else if (data_byte != ' '                       /* otherwise if we're overprinting a character */
          && data_byte != buffer [buffer_index])        /*   with a different character */
            buffer [buffer_index] = overprint_char;     /*     then substitute the overprint character */

        buffer_index++;                                 /* increment the buffer index */

        if (buffer_index % Zone_Size == 0)              /* if the zone is full */
            uptr->wait += dlyptr->zone_print;           /*   then add the zone printing delay */

        tprintf (lps_dev, TRACE_XFER, "Character %s sent to printer\n",
                 fmt_char (data_byte));
        }

    activate_unit (lps_unit);                           /* schedule DEMAND reassertion */
    }

return SCPE_OK;
}


/* Device reset routine.

   This routine is called for a RESET or RESET LPS command.  It is the
   simulation equivalent of applying or cycling power to the line printer.
*/

static t_stat lps_reset (DEVICE *dptr)
{
if (sim_switches & SWMASK ('P'))                        /* if this is a power-on reset */
    memcpy (times, &fast_times, sizeof fast_times);     /*    then reset the initial fast event delays */

if (power_state == power_on                             /* if the printer power is on */
  && lps_unit [0].flags & UNIT_ATT)                     /*   and paper is loaded */
    if (lps_unit [0].flags & UNIT_OFFLINE)              /*     then if the printer is offline */
        lps.status_word = ST_BUSY;                      /*       then set ready-but-busy status */
    else                                                /*     otherwise */
        lps.status_word = 0;                            /*       set ready-and-idle status */

else                                                    /* otherwise */
    lps.status_word = ST_NOT_READY | ST_BUSY;           /*   set not-ready-and-busy status */

sim_cancel (lps_unit);                                 /* deactivate the unit */

return SCPE_OK;
}



/* Interface local utility routines */


/* Activate a unit.

   The specified unit is added to the event queue with the delay specified by
   the unit wait field.


   Implementation notes:

    1. This routine may be called with wait = 0, which will expire immediately
       and enter the service routine with the next sim_process_event call.
       Activation is required in this case to allow the service routine to
       return an error code to stop the simulation.  If the service routine was
       called directly, any returned error would be lost.
*/

static void activate_unit (UNIT *uptr)
{
tprintf (lps_dev, TRACE_SERV, "Unit delay %u service scheduled\n",
         uptr->wait);

sim_activate (uptr, uptr->wait);                        /* activate the unit with the specified wait */

return;
}


/* Report a stream I/O error to the console.

   If a stream I/O error has been detected, this routine will print an error
   message to the simulation console and clear the stream's error indicator.
*/

static void report_error (FILE *stream)
{
cprintf ("%s simulator printer I/O error: %s\n",        /* report the error to the console */
         sim_name, strerror (errno));

clearerr (stream);                                      /* clear the error */

return;
}


/* Set the printer online or offline.

   This validation routine is called to set the printer online or offline.  The
   "value" parameter is UNIT_OFFLINE if the printer is going offline and is zero
   if the printer is going online.  This simulates toggling the ON LINE/OFF LINE
   switch on the printer.  The unit must be attached (i.e., paper must be
   loaded), before the printer may be set online.

   If a printer output operation is started via an STC instruction, and the
   printer is powered off, out of paper, or offline (i.e., DEMAND is denied),
   the Control and Command flip-flops will set, STROBE will be asserted, but no
   service event will be scheduled.  The I/O operation will "hang" at that point
   until the printer is powered on, supplied with paper (attached), or set
   online.  Placing the printer online with STROBE asserted will restart the
   operation by scheduling an event service.


   Implementation notes:

    1. This routine is called by SCP in response to a SET LPS OFF/ONLINE
       command.  It may also be called internally.  For the former, SCP takes
       care of setting or clearing the UNIT_OFFLINE flag.  For the latter, the
       flag must be set or cleared explicitly.  We do this unilaterally, as it's
       simpler, even though it is redundant in the former case.
*/

static t_stat lps_set_xline (UNIT *uptr, int32 value, char *cptr, void *desc)
{
if (value == UNIT_OFFLINE) {                            /* if setting the printer offline */
    lps.demand = FALSE;                                 /*   then deny DEMAND */
    lps.status_word |= ST_BUSY;                         /*     and set busy status */

    uptr->flags |= UNIT_OFFLINE;                        /* set the flag for local calls */
    }

else                                                    /* otherwise this is an online request */
    if (uptr->flags & UNIT_ATT) {                       /*   so if the paper is loaded */
        lps.status_word &= ~ST_BUSY;                    /*     then set idle status */
        lps.demand = TRUE;                              /*       and assert DEMAND */

        uptr->flags &= ~UNIT_OFFLINE;                   /* clear the flag for local calls */

        if (lps.strobe == TRUE)                         /* if STROBE is already asserted */
            sim_activate (uptr, 1);                     /*   then restart the output */
        }

    else                                                /* otherwise the printer is out of paper */
        return SCPE_NOFNC;                              /*   so it cannot be put online */

return SCPE_OK;
}


/* Attach the printer image file.

   The specified file is attached to the indicated unit.  This is the simulation
   equivalent of loading paper into the printer and setting the ON LINE switch.

   A new image file may be requested by giving the "-N" switch to the ATTACH
   command.  If an existing file is specified with "-N", it will be cleared; if
   specified without "-N", printer output will be appended to the end of the
   existing file content.  In all cases, the paper is positioned at the top of
   the form.

   Paper cannot be loaded into the printer unless the printer power is on.
*/

static t_stat lps_attach (UNIT *uptr, char *cptr)
{
t_stat result;

if (power_state == power_on) {                          /* if power is on */
    result = attach_unit (uptr, cptr);                  /*   then attach the printer image file for appending */

    if (result == SCPE_OK) {                            /* if the attach was successful */
        current_line = 1;                               /*   then reset the line counter to the top of form */
        buffer_index = 0;                               /*     and clear the buffer */
        overprint_index = 0;                            /*       and the overprint index too */

        tprintf (lps_dev, TRACE_CMD, "Printer paper loaded\n");

        lps.status_word &= ~ST_NOT_READY;               /* clear not-ready status */
        lps_set_xline (uptr, UNIT_ONLINE, NULL, NULL);  /*   and set the unit online */
        }
    }

else                                                    /* otherwise power is off */
    result = SCPE_NOFNC;                                /*   so paper loading is not allowed */

return result;
}


/* Detach the printer image file.

   The specified file is detached from the indicated unit.  This is the
   simulation equivalent of running out of paper or unloading the paper from the
   printer.  The out-of-paper condition cause a paper fault, and the printer
   goes offline.

   When the printer runs out of paper, the characters present in the buffer are
   printed and paper motion stops.  STROBE denies, and the printer goes offline.
*/

static t_stat lps_detach (UNIT *uptr)
{
t_stat result;

if (not (uptr->flags & UNIT_ATT))                       /* if the unit is not currently attached */
    return SCPE_UNATT;                                  /*   then report it */

else {                                                  /* otherwise */
    if (buffer_index > 0 || overprint_index > 0) {      /*   if the buffer is not empty */
        buffer [buffer_index++] = CR;                   /*     then tie it off */
        buffer [buffer_index++] = LF;                   /*       to complete the line */

        fwrite (buffer, sizeof buffer [0],              /* write the partial buffer */
                buffer_index, uptr->fileref);           /*   to the printer file */

        buffer_index = 0;                               /* clear the buffer index */
        overprint_index = 0;                            /*   and the overprint index */
        }

    current_line = 1;                                   /* reset the printer to TOF */
    sim_cancel (uptr);                                  /*   and terminate */
    lps.strobe = FALSE;                                 /*     any print action in progress */

    result = detach_unit (uptr);                        /* detach the unit */

    tprintf (lps_dev, TRACE_CMD, "Printer is out of paper\n");

    lps.status_word |= ST_NOT_READY;                    /* set not-ready status */
    lps_set_xline (uptr, UNIT_OFFLINE, NULL, NULL);     /*   and set the unit offline */

    return result;
    }
}


/* Set the device modes.

   This validation routine is entered with the "value" parameter set to one of
   the SIM_MODES values.  For device modes, the device flag implied by the value
   is set or cleared.  For unit modes, the unit action specified by the value is
   performed.  The character and descriptor pointers are not used.


   Implementation notes:

    1. When switching from diagnostic to printer mode, the current printer
       status is reestablished by calling the printer reset routine.  The
       command line switches are explicitly cleared so that the reset routine
       doesn't interpret them.

    2. Diagnostic mode cancels any pending event service.  This is necessary to
       prevent a power-returning service entry that might be in progress, as it
       would be misinterpreted if the printer is not connected to the status
       lines.
*/

static t_stat lps_set_mode (UNIT *uptr, int32 value, char *cptr, void *desc)
{
switch ((SIM_MODES) value) {                            /* dispatch the mode to set */

    case Fast_Time:                                     /* entering optimized timing mode */
        lps_dev.flags &= ~DEV_REALTIME;                 /*   so clear the real-time flag */
        break;


    case Real_Time:                                     /* entering realistic timing mode */
        lps_dev.flags |= DEV_REALTIME;                  /*   so set the real-time flag */
        break;


    case Printer:                                       /* entering printer mode */
        lps_dev.flags &= ~DEV_DIAG;                     /*   so clear the diagnostic flag */

        sim_switches = 0;                               /* reset the device to reestablish */
        lps_reset (&lps_dev);                           /*   the current printer status */
        break;


    case Diagnostic:                                    /* entering diagnostic mode */
        lps_dev.flags |= DEV_DIAG;                      /*   so set the diagnostic flag */
        lps_unit [0].wait = LOOPBACK_DELAY;             /*     and the loopback handshake delay */

        sim_cancel (uptr);                              /* cancel any output operation in progress */
        break;


    case TOF:                                           /* manual top-of-form request */
    case Step:                                          /* manual paper-step request */
        if (uptr->flags & UNIT_ATT                      /* if the printer paper is loaded */
          && uptr->flags & UNIT_OFFLINE                 /*   and the printer is offline */
          && power_state == power_on                    /*     and the power is on */
          && not (lps_dev.flags & DEV_DIAG)) {          /*       and the card is not looped back */
            lps.demand = TRUE;                          /*         then assert DEMAND */
            lps.strobe = TRUE;                          /*           and STROBE */

            if ((SIM_MODES) value == TOF)               /* if this is a top-of-form request */
                lps.output_word = FF;                   /*   then store the form-feed control */
            else                                        /* otherwise it is a paper-step request */
                lps.output_word = LF;                   /*   so store the line-feed control */

            tprintf (lps_dev, TRACE_CMD, "%s switch toggled\n",
                     ((SIM_MODES) value == TOF ? "Top of form" : "Paper step"));

            lps_service (uptr);                         /* call the printer service to move the paper */
            sim_cancel (uptr);                          /*   and then cancel the scheduled handshake */

            tprintf (lps_dev, TRACE_SERV, "Service canceled for manual action\n");
            }

        else                                            /* otherwise the printer mode is incompatible */
            return SCPE_NOFNC;                          /*   and the paper command is not allowed */
        break;


    case Clear:                                         /* master clear request */
        buffer_index = 0;                               /* clear the buffer index */
        overprint_index = 0;                            /*   and the overprint index */

        lps_set_xline (uptr, UNIT_OFFLINE, NULL, NULL); /* set the unit offline */
        break;
    }

return SCPE_OK;                                         /* mode changes always succeed */
}


/* Show the device modes.

   This display routine is called to show the device modes for the printer.  The
   output stream is passed in the "st" parameter, and the other parameters are
   ignored.  The power state, timing mode, and connection mode are printed.
*/

static t_stat lps_show_mode (FILE *st, UNIT *uptr, int32 value, void *desc)
{
fprintf (st, "power %s, %s timing, %s mode",            /* print the power, timing, and connection modes */
         (power_state == power_off ? "off" : "on"),
         (lps_dev.flags & DEV_REALTIME ? "realistic" : "fast"),
         (lps_dev.flags & DEV_DIAG ? "diagnostic" : "printer"));

return SCPE_OK;
}
